/****************************************************************
 SDI Security Services

 Copyright (C) 2004 by con terra GmbH

 Author: Arne Broering, University of Muenster

 Contact: Albert Remke, con terra GmbH, Martin-Luther-King-Weg 24,
 48155 Muenster, Germany, 52n@conterra.de

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 version 2 as published by the Free Software Foundation.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program (see gnu-gpl v2.txt); if not, write to
 the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA  02111-1307, USA or visit the web page of the Free
 Software Foundation, http://www.fsf.org.

 Created on: 20.09.2005
 ****************************************************************/

package org.n52.security.authentication;

import java.util.HashMap;
import java.util.Map;
import org.n52.security.common.util.StringUtils;
import org.opensaml.SAMLException;

/**
 * A CredentialFactory is used to relate authentication method urns to credential converters,
 * which create Credentail instances from plain strings.
 * <p/>
 * You can use the static "default factory" approach or manage your own mappings.
 * To add special credential-urn mappings, simple add apropriate converters to the factory.
 *
 * @author Arne Broering, Marko Reiprecht
 */
public class CredentialFactory {

    // ------------------------------ FIELDS ------------------------------

    private static CredentialFactory defaultFactory = new CredentialFactory();

    private Map m_authMethodsToConverterMap = new HashMap();

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Constructor creates a new instance.
     * The factory will contain the same settings like the original default factory.
     */
    public CredentialFactory() {
        // todo: better a dynamic registration or complete rework of the "authentication method" idea.
        add(PasswordAuthenticationMethod.URN, new UsernamePasswordCredentialConverter(true));
        add(SessionAuthenticationMethod.URN, new SessionIDCredentialConverter());
        add(WASAuthenticationMethod.URN, new SAMLCredentialConverter());
        add(SAMLResponseAuthenticationMethod.URN, new SAMLCredentialConverter());
    }

    /**
     * Method add.
     *
     * @param authenticationMethodUrn the authenticationMethodUrn.
     * @param converter               the converter.
     * @return CredentialFactory.
     */
    public CredentialFactory add(final String authenticationMethodUrn, final CredentialConverter converter) {
        if (authenticationMethodUrn == null) {
            throw new IllegalArgumentException("authenticationMethodUrn must not null");
        }
        if (converter == null) {
            throw new IllegalArgumentException("converter must not null");
        }
        m_authMethodsToConverterMap.put(authenticationMethodUrn, converter);
        return this;
    }

    // --------------------- GETTER / SETTER METHODS ---------------------

    /**
     * Method gets the defaultFactory.
     *
     * @return CredentialFactory.
     */
    public static CredentialFactory getDefaultFactory() {
        return defaultFactory;
    }

    /**
     * Method sets the defaultFactory.
     *
     * @param defaultFactory the defaultFactory.
     */
    public static void setDefaultFactory(final CredentialFactory defaultFactory) {
        CredentialFactory.defaultFactory = defaultFactory;
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Method clear.
     *
     * @return CredentialFactory.
     */
    public CredentialFactory clear() {
        m_authMethodsToConverterMap.clear();
        return this;
    }

    /**
     * Method createCredential.
     *
     * @param pMethod           the pMethod.
     * @param encodedCredential the encodedCredential.
     * @return Credential.
     */
    public Credential create(final AuthenticationMethod pMethod, final String encodedCredential) {
        return create(pMethod.getURN(), encodedCredential);
    }

    /**
     * Method createCredential.
     *
     * @param authenticationMethodUrn the authenticationMethodUrn.
     * @param encodedCredential       the encodedCredential.
     * @return Credential
     * @throws IllegalStateException if an authentication method was specified for which no
     *                               credential converter was registered
     */
    public Credential create(final String authenticationMethodUrn, final String encodedCredential) {
        CredentialConverter converter = get(authenticationMethodUrn);
        if (converter == null) {
            throw new IllegalStateException(
                    "no CredentialConverter is registered for authenticationMethodUrn <" + authenticationMethodUrn + ">");
        }
        return converter.convert(encodedCredential);
    }

    /**
     * Method createCredential.
     *
     * @param pMethod    the pMethod.
     * @param credential the encodedCredential.
     * @return Credential.
     */
    public String encode(final AuthenticationMethod pMethod, final Credential credential) {
        return encode(pMethod.getURN(), credential);
    }

    /**
     * Method createCredential.
     *
     * @param authenticationMethodUrn the authenticationMethodUrn.
     * @param credential              the encodedCredential.
     * @return Credential.
     */
    public String encode(final String authenticationMethodUrn, final Credential credential) {
        CredentialConverter converter = get(authenticationMethodUrn);
        if (converter == null) {
            throw new IllegalStateException(
                    "no CredentialConverter is registered for authenticationMethodUrn <" + authenticationMethodUrn + ">");
        }
        return converter.convert(credential);
    }

    /**
     * Method get.
     *
     * @param authenticationMethodUrn the authenticationMethodUrn.
     * @return CredentialConverter.
     */
    public CredentialConverter get(final String authenticationMethodUrn) {
        if (authenticationMethodUrn == null) {
            throw new IllegalArgumentException("authenticationMethodUrn must not null");
        }
        return (CredentialConverter) m_authMethodsToConverterMap.get(authenticationMethodUrn);
    }

    /**
     * Method remove.
     *
     * @param authenticationMethodUrn the authenticationMethodUrn.
     * @return CredentialFactory.
     */
    public CredentialFactory remove(final String authenticationMethodUrn) {
        if (authenticationMethodUrn == null) {
            throw new IllegalArgumentException("authenticationMethodUrn must not null");
        }
        m_authMethodsToConverterMap.remove(authenticationMethodUrn);
        return this;
    }

    // -------------------------- INNER CLASSES --------------------------

    /**
     * A CredentialConverter is responsible to convert a plain string in a concrete credential instance and vise versa.
     *
     * @author Marko Reiprecht
     *         Created on 18.07.2008
     * @version $Revision: 1238 $
     */
    public interface CredentialConverter {

        /**
         * Method converts a plain string in a credential instance.
         *
         * @param encodedCredential the encodedCredential.
         * @return Credential.
         */
        Credential convert(String encodedCredential);

        /**
         * Method converts a credential to a plain string.
         *
         * @param credential the credential.
         * @return String.
         */
        String convert(Credential credential);
    }

    /**
     * Class Base64UsernamePasswordCredentialConverter.
     *
     * @author Marko Reiprecht
     *         Created on 18.07.2008
     * @version $Revision: 1238 $
     */
    public static class UsernamePasswordCredentialConverter implements CredentialConverter {

        private boolean m_base64Credentials = false;

        /**
         * Constructor creates a new instance.
         *
         * @param base64 the base64.
         */
        public UsernamePasswordCredentialConverter(final boolean base64) {
            m_base64Credentials = base64;
        }

        /**
         * Method converts a plain string in a credential instance.
         *
         * @param encodedCredential the encodedCredential.
         * @return Credential.
         */
        public Credential convert(final String encodedCredential) {
            // expected 'base 64 username, base 64 passwd'
            if (encodedCredential == null) {
                throw new IllegalArgumentException("the encoded credential must not null");
            }
            // if password is empty, then the splitted string has the length 1
            String[] splittedString = encodedCredential.split(",");
            if (splittedString.length != 1 && splittedString.length != 2) {
                throw new IllegalArgumentException(
                        "the encoded credential must contain two values or at lease one comma");
            }
            String username = splittedString[0];
            String pw = splittedString.length == 2 ? splittedString[1] : "";
            if (m_base64Credentials) {
                username = StringUtils.decodeBase64(username);
                pw = StringUtils.decodeBase64(pw);
            }
            return new UsernamePasswordCredential(username, pw);
        }

        /**
         * Method converts a credential to a plain string.
         *
         * @param credential the credential.
         * @return String.
         */
        public String convert(final Credential credential) {
            if (!(credential instanceof UsernamePasswordCredential)) {
                throw new IllegalArgumentException("credential must have type UsernamePasswordCredential");
            }
            UsernamePasswordCredential usercred = (UsernamePasswordCredential) credential;
            String name = usercred.getUsername();
            String pw = new String(usercred.getPassword());
            if (m_base64Credentials) {
                name = StringUtils.encodeBase64(name);
                pw = StringUtils.encodeBase64(pw);
            }
            return new StringBuffer()
                    .append(name)
                    .append(",")
                    .append(pw)
                    .toString();
        }
    }

    /**
     * Class SessionIDCredentialConverter.
     *
     * @author Marko Reiprecht
     *         Created on 18.07.2008
     * @version $Revision: 1238 $
     */
    public static class SessionIDCredentialConverter implements CredentialConverter {

        /**
         * Method converts a plain string in a credential instance.
         *
         * @param encodedCredential the encodedCredential.
         * @return Credential.
         */
        public Credential convert(final String encodedCredential) {
            // expected 'sessionid'
            if (encodedCredential == null) {
                throw new IllegalArgumentException("the encoded credential must not null");
            }
            if (encodedCredential.length() == 0) {
                throw new IllegalArgumentException("the encoded credential must have a length > 0");
            }
            return new SessionIDCredential(encodedCredential);
        }

        /**
         * Method converts a credential to a plain string.
         *
         * @param credential the credential.
         * @return String.
         */
        public String convert(final Credential credential) {
            if (!(credential instanceof SessionIDCredential)) {
                throw new IllegalArgumentException("credential must have type SessionIDCredential");
            }
            return ((SessionIDCredential) credential).getSessionId();
        }
    }

    /**
     * Class SAMLCredentialConverter.
     *
     * @author Marko Reiprecht
     *         Created on 18.07.2008
     * @version $Revision: 1238 $
     */
    public static class SAMLCredentialConverter implements CredentialConverter {

        /**
         * Method converts a plain string in a credential instance.
         *
         * @param encodedCredential the encodedCredential.
         * @return Credential.
         */
        public Credential convert(final String encodedCredential) {
            // expected 'base 64  encoded saml response'
            if (encodedCredential == null) {
                throw new IllegalArgumentException("the encoded credential must not null");
            }
            if (encodedCredential.length() == 0) {
                throw new IllegalArgumentException("the encoded credential must have a length > 0");
            }
            try {
                return new SAMLCredential(new SAMLTicket(StringUtils.decodeBase64(encodedCredential)));
            } catch (SAMLException e) {
                throw new IllegalArgumentException("encoded credential can not be converted to a SAML-Response :" + e);
            }
        }

        /**
         * Method converts a credential to a plain string.
         *
         * @param credential the credential.
         * @return String.
         */
        public String convert(final Credential credential) {
            if (!(credential instanceof SAMLCredential)) {
                throw new IllegalArgumentException("credential must have type SAMLCredential");
            }
            SAMLCredential samlCred = (SAMLCredential) credential;
            return samlCred.getSAMLTicket().asBase64String();
        }
    }
}
