/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 18.07.2008
 * Modified: $Date $
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.authentication;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Date;
import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.Configuration;
import javax.security.auth.login.CredentialExpiredException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.n52.security.authentication.callbacks.NoCredentialsCallbackHandler;

/**
 * Class JAASAuthenticationService implements the AuthenticationService interface
 * by using the JAAS api.
 * <p/>
 * This AuthenticationService implementation is the default implementation used within 52n-security.
 * If you create an instance it uses the default JAAS Configuration
 * ({@link javax.security.auth.login.Configuration#getConfiguration()} and an empty application name to perform logins.
 * <p/>
 * You can use the setter methods for the applicationname and the Configuration to change the behavior of this class.
 *
 * @author Marko Reiprecht
 *         Created on 18.07.2008
 * @version $Revision: $
 */
public class JAASAuthenticationService implements AuthenticationService {

    // ------------------------------ FIELDS ------------------------------

    /**
     * The Logger.
     */
    private static final Log LOG = LogFactory.getLog(JAASAuthenticationService.class);

    private Configuration m_configuration = null;
    private String m_applicationName = "";

    // --------------------- GETTER / SETTER METHODS ---------------------

    /**
     * Constructor creates a new instance.
     */
    public JAASAuthenticationService() {
        try {
            m_configuration = Configuration.getConfiguration();
        } catch (java.lang.SecurityException ex) {
            if (LOG.isInfoEnabled()) { LOG.info("No JAAS-Default configuration found"); }
        }
    }

    /**
     * Method gets the applicationName.
     *
     * @return String.
     * @see {@link javax.security.auth.login.Configuration} for details about the application name.
     */
    public String getApplicationName() {
        return m_applicationName;
    }

    /**
     * Method sets the applicationName.
     *
     * @param applicationName the applicationName.
     * @see {@link javax.security.auth.login.Configuration} for details about the application name.
     */
    public void setApplicationName(final String applicationName) {
        m_applicationName = applicationName;
    }

    /**
     * Method gets the configuration.
     *
     * @return Configuration.
     */
    public Configuration getConfiguration() {
        if (m_configuration == null) {
            throw new IllegalStateException("configuration instance not set on JAASAuthenticationService");
        }
        return m_configuration;
    }

    /**
     * Method sets the configuration.
     *
     * @param configuration the configuration.
     */
    public void setConfiguration(final Configuration configuration) {
        m_configuration = configuration;
    }

    // ------------------------ INTERFACE METHODS ------------------------


    // --------------------- Interface AuthenticationService ---------------------

    /**
     * @see AuthenticationService#createAuthenticationContext()
     */
    public AuthenticationContext createAuthenticationContext() {
        return new JAASAuthenticationContext(new Subject());
    }

    /**
     * @see AuthenticationService#login()
     */
    public AuthenticationContext login() throws AuthenticationException {
        return login(null, null);
    }

    /**
     * @see AuthenticationService#login(CallbackHandler)
     */
    public AuthenticationContext login(final CallbackHandler handler) throws AuthenticationException {
        return login(handler, null);
    }

    /**
     * @see AuthenticationService#login(CallbackHandler, AuthenticationContext)
     */
    public AuthenticationContext login(final CallbackHandler handler, final AuthenticationContext ctx)
            throws AuthenticationException {
        JAASAuthenticationContext authContext = (JAASAuthenticationContext) (ctx == null
                                                                             ? createAuthenticationContext()
                                                                             : ctx);

        CallbackHandler callbackHandler = handler == null ? new NoCredentialsCallbackHandler() : handler;
        try {
            LoginContext loginContext = createLoginContext(getApplicationName(),
                                                           authContext.getSubject(),
                                                           handler,
                                                           getConfiguration());
            authContext.setLoginContext(loginContext);

            loginContext.login();
            authContext.setAuthenticated(true);
            authContext.setAuthenticationTime(new Date());
        } catch (LoginException e) {
            authContext.setAuthenticated(false);
            authContext.setAuthenticationTime(null);
            throw wrap(e);
        }
        return authContext;
    }

    /**
     * @see AuthenticationService#logout(AuthenticationContext)
     */
    public AuthenticationContext logout(final AuthenticationContext ctx) throws AuthenticationException {
        if (ctx == null) {
            throw new IllegalArgumentException("authenticationContext must not null");
        }
        JAASAuthenticationContext authContext = (JAASAuthenticationContext) ctx;
        if (authContext.isAuthenticated()) {
            LoginContext loginContext = authContext.getLoginContext();
            if (loginContext != null) {// possible if authentication context was serialized
                try {
                    loginContext.logout();
                    authContext.setAuthenticated(false);
                    authContext.setAuthenticationTime(null);
                } catch (LoginException e) {
                    throw wrap(e);
                }
            }
        }
        return authContext;
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Method creates a LoginContext.
     *
     * @param applicationName the applicationName.
     * @param subject         the subject.
     * @param handler         the handler.
     * @param configuration   the configuration.
     * @return LoginContext.
     */
    protected LoginContext createLoginContext(final String applicationName,
                                              final Subject subject,
                                              final CallbackHandler handler,
                                              final Configuration configuration) {
        try {
            try {
                // here a little reflection because jdk 1.5 provides non static configuration binding
                Class[] params = {String.class, Subject.class, CallbackHandler.class, Configuration.class};
                Constructor constr = LoginContext.class.getConstructor(params);
                return (LoginContext) constr.newInstance(
                        new Object[]{applicationName, subject, handler, configuration});
            } catch (NoSuchMethodError e) {
                try {
                    // use jdk 1.4 style
                    // we have to save the static configuration
                    synchronized (Configuration.class) {
                        Configuration staticConf = Configuration.getConfiguration();
                        Configuration.setConfiguration(configuration);
                        try {
                            return new LoginContext(applicationName, subject, handler);
                        } finally {
                            Configuration.setConfiguration(staticConf);
                        }
                    }
                } catch (LoginException ex) {
                    throw wrap(ex);
                }
            }
        } catch (InvocationTargetException e) {
            if (e.getTargetException() instanceof LoginException) {
                throw wrap((LoginException) e.getTargetException());
            } else if (e.getTargetException() instanceof AuthenticationException) {
                throw (AuthenticationException) e.getTargetException();
            } else {
                throw new AuthenticationServiceNotAvailableException("unclassified exception thrown", e);
            }
        } catch (Exception e) {
            throw new AuthenticationServiceNotAvailableException("login context can not be created", e);
        }
    }

    /**
     * Method converts a LoginException in a AuthenticationException.
     *
     * @param ex the ex.
     * @return AuthenticationException.
     */
    protected AuthenticationException wrap(final LoginException ex) {
        if (ex instanceof javax.security.auth.login.AccountExpiredException) {
            return new AccountExpiredException("account expired", ex);
        } else if (ex instanceof CredentialExpiredException) {
            return new CredentialsExpiredException("credentials expired", ex);
        } else if (ex instanceof FailedLoginException) {
            return new AuthenticationFailedException("authentication failed", ex);
        } else {
            return new AuthenticationServiceNotAvailableException("unspecified login exception thrown", ex);
        }
    }

}
