/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 18.07.2008
 * Modified: $Date $
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.authentication.callbacks;

import java.io.IOException;
import java.util.Map;
import java.util.WeakHashMap;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.callback.UnsupportedCallbackException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.n52.security.authentication.Credential;
import org.n52.security.authentication.UsernamePasswordCredential;

/**
 * Class CredentialsCallbackHandler reacts on CredentialCallbacks.
 * <p/>
 * A user can add credentials to this handler and if login modules ask for the classes,
 * the credentials are provided by this handler.
 *
 * @author Marko Reiprecht
 *         Created on 18.07.2008
 * @version $Revision: $
 */
public class CredentialsCallbackHandler implements CallbackHandler {

    /**
     * The Logger.
     */
    private static final Log LOG = LogFactory.getLog(CredentialsCallbackHandler.class);

    private Map m_credentials = new WeakHashMap();

    /**
     * Method handles CredentialCallbacks, special behavior is implemented for NameCallbacks and PasswordCallbacks,
     * both are mapped to UsernamePasswordCredential.class callbacks.
     *
     * @param callbacks the callbacks.
     * @throws IOException                  if error occurs.
     * @throws UnsupportedCallbackException if error occurs.
     */
    public void handle(final Callback[] callbacks) throws IOException, UnsupportedCallbackException {
        for (int i = 0, l = callbacks.length; i < l; ++i) {
            Callback callback = callbacks[i];
            if (callback instanceof CredentialCallback) {
                CredentialCallback credCallback = ((CredentialCallback) callback);
                Credential credential = lookUp(credCallback.getRequiredCredentialClass());
                if (credential != null) {
                    credCallback.setCredential(credential);
                } else {
                    if (LOG.isWarnEnabled()) {
                        LOG.warn(
                                "no credential of class <" + credCallback.getRequiredCredentialClass() + "> available");
                    }
                }
            } else if (callback instanceof NameCallback) {
                UsernamePasswordCredential cred = (UsernamePasswordCredential) lookUp(UsernamePasswordCredential.class);
                if (cred != null) {
                    NameCallback namecallback = (NameCallback) callback;
                    namecallback.setName(cred.getUsername());
                } else {
                    if (LOG.isWarnEnabled()) {
                        LOG.warn("username requested, but no UsernamePasswordCredential available");
                    }
                }
            } else if (callback instanceof PasswordCallback) {
                UsernamePasswordCredential cred = (UsernamePasswordCredential) lookUp(UsernamePasswordCredential.class);
                if (cred != null) {
                    PasswordCallback pwcallback = (PasswordCallback) callback;
                    pwcallback.setPassword(cred.getPassword());
                } else {
                    if (LOG.isWarnEnabled()) {
                        LOG.warn("username requested, but no UsernamePasswordCredential available");
                    }
                }
            } else {
                if (LOG.isWarnEnabled()) {
                    LOG.warn("can not handle callback of type <" + callback.getClass() + "> it is ignored");
                }
            }
        }
    }

    /**
     * Method lookUp searches for a credential of given type.
     *
     * @param credentialClass the credentialClass.
     * @return Credential.
     */
    private Credential lookUp(final Class credentialClass) {
        return (Credential) m_credentials.get(credentialClass);
    }

    /**
     * Method adds a new credential to the available set.
     * Credentials of the same type replaces the first added.
     *
     * @param credential the credential.
     * @return CredentialsCallbackHandler.
     */
    public CredentialsCallbackHandler add(final Credential credential) {
        if (credential == null) {
            throw new IllegalArgumentException("credential must not null");
        }
        m_credentials.put(credential.getClass(), credential);
        return this;
    }
}
