/*******************************************************************************
 * Copyright  2007 52North Initiative for Geospatial Open Source Software GmbH
 *
 * Author: Marko Reiprecht, con terra GmbH
 *
 * Contact: Andreas Wytzisk, 52North Initiative for Geospatial Open Source
 * Software GmbH, Martin-Luther-King-Weg 24, 48155 Muenster, Germany,
 * info@52north.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; even without the implied WARRANTY OF MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundations web page, http://www.fsf.org.
 *
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.authentication.config;

import java.util.Collections;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import javax.security.auth.login.AppConfigurationEntry;
import javax.security.auth.login.Configuration;

/**
 * Class BeanConfiguration represents a JAAS-Configuration, which is written in a Pojo-Style.
 * <p/>
 * It provides methods for simple setting/getting the AppConfigurationEntry[] instances.
 * It does not implement any kind of external configuration.
 * It extends {@link ConfigurationDecorator},
 * therewith it is possible to generate a chain of Configuration-instances.
 *
 * @author Marko Reiprecht
 *         Created on 10.09.2007
 * @version $Revision: $
 */
public class BeanConfiguration extends ConfigurationDecorator {

    /**
     * Contains as keys Strings and as values AppConfigurationEntry[] arrays.
     */
    private Map m_appConfigurationEntries = new Hashtable();

    /**
     * Constructor creates a empty BeanConfiguration instance.
     */
    public BeanConfiguration() {
        super(null);
    }


    /**
     * Constructor creates a empty BeanConfiguration instance, which delegates to
     * the given Configuration instance.
     *
     * @param configuration the configuration to delegate to.
     * @see ConfigurationDecorator#ConfigurationDecorator(Configuration)
     */
    public BeanConfiguration(final Configuration configuration) {
        super(configuration);
    }

    /**
     * Constructor creates a empty BeanConfiguration instance, which delegates to
     * the given Configuration instance.
     *
     * @param configuration of type Configuration, can be null.
     * @param weak          of type boolean, if true only a {@link java.lang.ref.WeakReference}
     *                      is stored to the given Configuration.
     * @see ConfigurationDecorator#ConfigurationDecorator(Configuration,boolean)
     */
    public BeanConfiguration(final Configuration configuration,
                             final boolean weak) {
        super(configuration, weak);
    }

    /**
     * Constructor creates a BeanConfiguration instance, which only delegates to
     * the given Configuration instance if no application entry is found in the given map.
     *
     * @param configuration           of type Configuration, can be null.
     * @param weak                    of type boolean, if true only a {@link java.lang.ref.WeakReference}
     *                                is stored to the given Configuration.
     * @param appConfigurationEntries a Map of type Map&lt;String,AppConfigurationEntry[]&gt;
     * @see ConfigurationDecorator#ConfigurationDecorator(Configuration,boolean)
     */
    public BeanConfiguration(final Configuration configuration, final boolean weak,
                             final Map appConfigurationEntries) {
        super(configuration, weak);
        setAppConfigurationEntries(appConfigurationEntries);
    }


    /**
     * Method returns the appConfigurationEntries managed by this BeanConfiguration object.
     * The returned map is unmodifiable and only the entries directly visible to this
     * instance are returned. Entries managed elsewere in the configuration chain are not returned.
     *
     * @return the appConfigurationEntries (type Map&lt;String,AppConfigurationEntry[]&gt;).
     */
    public Map getAppConfigurationEntries() {
        return Collections.unmodifiableMap(m_appConfigurationEntries);
    }

    /**
     * Method sets the appConfigurationEntries of this BeanConfiguration object.
     * The entries of the given map are copied to the internal map.
     *
     * @param configurationEntries the appConfigurationEntries managed by this BeanConfiguration object.
     */
    public void setAppConfigurationEntries(final Map configurationEntries) {
        // clean old configuration
        m_appConfigurationEntries.clear();
        // iterate through map and check for correct types and add to internal map
        Iterator lIt = configurationEntries.entrySet().iterator();
        while (lIt.hasNext()) {
            Map.Entry mapentry = (Map.Entry) lIt.next();
            setAppConfigurationEntry(String.valueOf(mapentry.getKey()),
                                     (AppConfigurationEntry[]) mapentry.getValue());
        }
    }

    /**
     * Method sets the AppConfigurationEntries of a special application name.
     * The given entry is copied to the internal map.
     * A entry with the same name will be overwritten.
     *
     * @param applicationName  of type String must not null.
     * @param applicationEntry of type AppConfigurationEntry[] must not null.
     */
    public void setAppConfigurationEntry(final String applicationName,
                                         final AppConfigurationEntry[] applicationEntry) {
        if (applicationName == null) { throw new NullPointerException("applicationName"); }
        if (applicationEntry == null) { throw new NullPointerException("applicationEntries"); }

        if (applicationEntry.length != 0) {
            m_appConfigurationEntries.put(applicationName, createCopy(applicationEntry));
        } else {
            removeAppConfigurationEntry(applicationName);
        }
    }

    /**
     * Method removes all AppConfigurationEntries of a special application name.
     * Only entries managed(visible) by this Configuration instance are removed.
     *
     * @param applicationName of type String must not null.
     */
    public void removeAppConfigurationEntry(final String applicationName) {
        if (applicationName == null) { throw new NullPointerException("applicationName"); }
        m_appConfigurationEntries.remove(applicationName);
    }

    /**
     * Returns the AppConfiguraionEntries for a special applicationName.
     * First it looks in the entries managed by this object, if no entries for the applicationName
     * are found the wrapped Configuration object is asked.
     *
     * @param applicationName of type  String must not null.
     * @return AppConfigurationEntry[] indexed by applicationName.
     * @see ConfigurationDecorator#getAppConfigurationEntry(String)
     */
    public AppConfigurationEntry[] getAppConfigurationEntry(final String applicationName) {
        AppConfigurationEntry[] lEntry =
                (AppConfigurationEntry[]) m_appConfigurationEntries.get(applicationName);
        return lEntry == null ? super.getAppConfigurationEntry(applicationName) : lEntry;
    }

    /**
     * Method creates a 'deep' copy of the source array and stores the new created objects in target.
     * Asserts source.length <= target.length.
     *
     * @param source of type AppConfigurationEntry[]
     * @param target of type AppConfigurationEntry[]
     * @return AppConfigurationEntry[]
     */
    private AppConfigurationEntry[] createCopy(final AppConfigurationEntry[] source,
                                               final AppConfigurationEntry[] target) {
        for (int i = 0, l = source.length; i < l; ++i) {
            AppConfigurationEntry entry = source[i];
            target[i] = new AppConfigurationEntry(entry.getLoginModuleName(),
                                                  entry.getControlFlag(),
                                                  entry.getOptions());
        }
        return target;
    }

    /**
     * Method creates a 'deep' copy of the source array.
     *
     * @param source of type AppConfigurationEntry[]
     * @return AppConfigurationEntry[]
     */
    private AppConfigurationEntry[] createCopy(final AppConfigurationEntry[] source) {
        AppConfigurationEntry[] copy = new AppConfigurationEntry[source.length];
        return createCopy(source, copy);
    }
}
