/*******************************************************************************
 * Copyright  2007 52North Initiative for Geospatial Open Source Software GmbH
 *
 * Author: Marko Reiprecht, con terra GmbH
 *
 * Contact: Andreas Wytzisk, 52North Initiative for Geospatial Open Source
 * Software GmbH, Martin-Luther-King-Weg 24, 48155 Muenster, Germany,
 * info@52north.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; even without the implied WARRANTY OF MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundations web page, http://www.fsf.org.
 *
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.authentication.config;

import java.lang.ref.WeakReference;
import javax.security.auth.login.AppConfigurationEntry;
import javax.security.auth.login.Configuration;

/**
 * Class ConfigurationDecorator represents a decorator for a JAAS-Configuration.
 * <p/>
 * It does not implement any kind of external configuration.
 * It wrapps(decorates) another Configuration class and delegates all methods to this instance.
 * It can be used to create chains of configurations
 * or to change the behavior of the decorated configuration.
 * <p/>
 * It also provides a 'weak' mode. In this mode only a weak reference to the wrapped configuration
 * is stored and therewith the decorated instance can be garbaged.
 * This mode can be used for advanced application requirements.
 *
 * @author Marko Reiprecht
 *         Created on 10.09.2007
 * @version $Revision: $
 */
public class ConfigurationDecorator extends Configuration {

    /**
     * Field m_configuration  holds a strong reference to a configuration,
     * this is null if only the weak refernce is used.
     */
    private Configuration m_configuration;
    /**
     * Field m_configurationReference holds a weak reference to a configuration.
     */
    private WeakReference m_configurationReference;


    /**
     * Creates a new AbstractConfigurationDecorator instance
     * delegating to the given Configuration instance.
     *
     * @param configuration of type Configuration
     */
    public ConfigurationDecorator(final Configuration configuration) {
        this(configuration, false);
    }


    /**
     * Creates a new AbstractConfigurationDecorator instance
     * delegating to the given Configuration instance.
     *
     * @param configuration of type Configuration, can be null.
     * @param weak          of type boolean, if true only a {@link java.lang.ref.WeakReference}
     *                      is stored to the given Configuration.
     */
    public ConfigurationDecorator(final Configuration configuration,
                                           final boolean weak) {
        setDelegationConfiguration(configuration, weak);
    }

    /**
     * Returns the underlying configuration object.
     *
     * @return the config type Configuration or null if garbaged or not set.
     */
    public synchronized Configuration getDelegationConfiguration() {
        return (Configuration) m_configurationReference.get();
    }


    /**
     * Method sets the underlying configuration object.
     *
     * @param configuration of type Configuration, this can be null to unset the delegation configuration.
     * @param weak          of type boolean, if true only a {@link java.lang.ref.WeakReference}
     *                      is stored to the given Configuration.
     */
    public synchronized void setDelegationConfiguration(final Configuration configuration,
                                                        final boolean weak) {
        if (!weak) {
            m_configuration = configuration;
        }
        m_configurationReference = new WeakReference(configuration);
    }

    /**
     * Method sets the underlying configuration object.
     *
     * @param configuration of type Configuration.
     */
    public void setDelegationConfiguration(final Configuration configuration) {
        setDelegationConfiguration(configuration, false);
    }

    /**
     * Refreshs the configuration by delegating to the referenced Configuration.
     *
     * @see javax.security.auth.login.Configuration#refresh()
     */
    public void refresh() {
        Configuration lConfig = getDelegationConfiguration();
        if (lConfig != null) {
            lConfig.refresh();
        }
    }

    /**
     * Method getAppConfigurationEntry returns the AppConfigurationEntries for an application name
     * by delegateing to the referenced configuration.
     *
     * @param applicationName of type String a index name to the configuration of the calling application.
     * @return AppConfigurationEntry[] the AppConfigurationEntries configured for the application.
     * @see javax.security.auth.login.Configuration#getAppConfigurationEntry(String)
     */
    public AppConfigurationEntry[] getAppConfigurationEntry(final String applicationName) {
        Configuration lConfig = getDelegationConfiguration();
        if (lConfig != null) {
            return lConfig.getAppConfigurationEntry(applicationName);
        }
        return null;
    }
}
