/*******************************************************************************
 * Copyright  2007 52North Initiative for Geospatial Open Source Software GmbH
 *
 * Author: Marko Reiprecht, con terra GmbH
 *
 * Contact: Andreas Wytzisk, 52North Initiative for Geospatial Open Source
 * Software GmbH, Martin-Luther-King-Weg 24, 48155 Muenster, Germany,
 * info@52north.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; even without the implied WARRANTY OF MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundations web page, http://www.fsf.org.
 *
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.authentication.loginmodule;

import java.io.IOException;
import java.io.Serializable;
import java.security.Principal;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import javax.security.auth.DestroyFailedException;
import javax.security.auth.Destroyable;
import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.TextOutputCallback;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Class AbstractLoginModule is a abstract base class for login modules.
 * <p/>
 * It implements the methods
 * {@link javax.security.auth.spi.LoginModule#initialize(javax.security.auth.Subject,javax.security.auth.callback.CallbackHandler,java.util.Map,java.util.Map)},
 * {@link javax.security.auth.spi.LoginModule#commit()},
 * {@link javax.security.auth.spi.LoginModule#abort()} and
 * {@link javax.security.auth.spi.LoginModule#logout()} in an open manner to support a simpler
 * and error free implementation of concrete login modules.
 * <p/>
 * Sub classes have to implement the methods {@link #initialize()},
 * {@link #clearAuthenticationState()}, {@link #performLogin()} and {@link #prepareCommitState()}.
 * <p/>
 * Within the initialize method the login module can check given options and the shared state
 * for context information.
 * <p/>
 * In the performLogin method the login module performs the login,
 * which means that it checks the authentication credentials.
 * <p/>
 * The prepareCommitState method is invoked during the commit phase of the authentication. Therein
 * a login module can store all principals and credentials for the user in the
 * local principal and credentail sets ({@link #addPrincipal(java.security.Principal)},
 * ({@link #addPrivateCredential(Object)} and ({@link #addPublicCredential(Object)}).
 * These will automatically applied to the user and removed in the {@link #logout()}
 * and {@link #abort()} methods.
 * <p/>
 * The clearAuthenticationState method is invoked to tell a login module to release any authentication
 * information optained from the user.
 * <p/>
 * This class implements Serializable to make it possible to store LoginContexts for a subject for a longer time.
 * Not all fields can be serialized e.g. the CallbackHandler is omitted, therewith it makes
 * only sense to serialize authenticated login contexts, e.g. for later logout in a web application.
 * Successfull serialization requires also serializable credentials and principals.
 *
 * @author Marko Reiprecht
 *         Created on 15.09.2007
 * @version $Revision: $
 * @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jaas/JAASLMDevGuide.html">
 *      LoginModule Developer's Guide
 *      </a>
 */
public abstract class AbstractLoginModule implements LoginModule, Serializable {

    // ------------------------------ FIELDS ------------------------------

    /**
     * The Logger.
     */
    private static final Log LOG = LogFactory.getLog(AbstractLoginModule.class);
    /**
     * The serialVersionUID.
     *
     * @see java.io.Serializable
     */
    private static final long serialVersionUID = 9133908238663307934L;

    /**
     * Field m_subject holds the Subject.
     *
     * @see #getSubject()
     */
    protected Subject m_subject;
    /**
     * Field m_callbackHandler  holds the CallbackHandler.
     *
     * @see #getCallbackHandler()
     * @see #isCallbackHandlerSet()
     */
    protected transient CallbackHandler m_callbackHandler;
    /**
     * Field m_sharedState  holds the shared state.
     *
     * @see #getSharedState()
     */
    protected SharedState m_sharedState;
    /**
     * Field m_options  holds the options of the login module.
     *
     * @see #getOptions()
     */
    protected Options m_options;

    /**
     * Field m_principals contains the principals,
     * which shall or were applied to the subject.
     */
    protected Set m_principals;
    /**
     * Field m_publicCredentials  contains the public credentials,
     * which are created by this login module and shall or were applied to the subject.
     */
    protected Set m_publicCredentials;
    /**
     * Field m_privateCredentials  contains the private credentials,
     * which are created by this login module and shall or were applied to the subject.
     */
    protected Set m_privateCredentials;

    /**
     * Field m_loginSucceeded is a flag wich indicates a successfull execution of the login method.
     */
    protected boolean m_loginSucceeded;
    /**
     * Field m_commitSucceeded  is a flat wich indicates a successfull execution of the commit method.
     */
    protected boolean m_commitSucceeded;

    // --------------------- GETTER / SETTER METHODS ---------------------

    /**
     * Method gets the callbackHandler.
     * It can be null if the application does not set one.
     *
     * @return CallbackHandler | null
     */
    protected CallbackHandler getCallbackHandler() {
        return m_callbackHandler;
    }

    /**
     * Method gets the options of this AbstractLoginModule object.
     *
     * @return Options.
     */
    protected Options getOptions() {
        return m_options;
    }

    /**
     * Method gets the principals created during the login method.
     *
     * @return Set of {@link java.security.Principal} classes.
     */
    protected Set getPrincipals() {
        return m_principals;
    }

    /**
     * Method gets the privateCredentials created during the login method.
     *
     * @return Set of credentials -> type is unknown and specific to the login module.
     */
    protected Set getPrivateCredentials() {
        return m_privateCredentials;
    }

    /**
     * Method gets the publicCredentials created during the login method.
     *
     * @return Set of credentials -> type is unknown and specific to the login module.
     */
    protected Set getPublicCredentials() {
        return m_publicCredentials;
    }

    /**
     * Method gets the sharedState of this login context.
     *
     * @return SharedState.
     */
    protected SharedState getSharedState() {
        return m_sharedState;
    }

    /**
     * Method gets the subject of this login context.
     *
     * @return Subject.
     */
    protected Subject getSubject() {
        return m_subject;
    }

    /**
     * Method gets the commitSucceeded flag.
     * This is must be true only if the commit method was executed successfull.
     *
     * @return boolean.
     */
    protected boolean isCommitSucceeded() {
        return m_commitSucceeded;
    }

    /**
     * Method sets the commit succeeded state of the login module.
     * This must be called with a value of true if the commit method of a subclass succeeded.
     *
     * @param commitSucceeded (true|false)
     */
    protected void setCommitSucceeded(final boolean commitSucceeded) {
        m_commitSucceeded = commitSucceeded;
    }

    /**
     * Method gets the loginSucceeded flag.
     * This must be true only if the login method was executed successfull.
     *
     * @return boolean.
     */
    protected boolean isLoginSucceeded() {
        return m_loginSucceeded;
    }

    /**
     * Method sets the login succeeded state of the login module.
     * This must be called with a m_value of true if the login method of a subclass succeeded.
     *
     * @param loginSucceeded (true|false).
     */
    protected void setLoginSucceeded(final boolean loginSucceeded) {
        m_loginSucceeded = loginSucceeded;
    }

    // ------------------------ INTERFACE METHODS ------------------------


    // --------------------- Interface LoginModule ---------------------

    /**
     * Method aborts the authentication process.
     *
     * @return boolean.
     * @throws LoginException if error occurs.
     * @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jaas/JAASLMDevGuide.html#abort">
     *      Description of the abort method in the LoginModule Developer's Guide
     *      </a>
     */
    public boolean abort() throws LoginException {
        if (LOG.isDebugEnabled()) { LOG.debug("AbstractLoginModule.abort"); }
        if (isLoginSucceeded()) {
            /*
            COMMIT \ ABORT (assumes LOGIN method succeeded)
                     \   SUCCESS         FAILURE
                    +--------------+-----------------+
                    |              |                 |
            SUCCESS | return TRUE  | throw EXCEPTION |
                    |              |                 |
                    +--------------+-----------------+
                    |              |                 |
            FAILURE | return TRUE  | throw EXCEPTION |
                    |              |                 |
                    +--------------+-----------------+
        */
            if (isCommitSucceeded()) {
                return logout();
            } else {
                // state was not applied, simply clear internal state
                return clearInternalState();
            }
        } else {
            /*
            COMMIT \ ABORT (assumes LOGIN method failed)
                     \   SUCCESS         FAILURE
                    +--------------+-----------------+
                    |              |                 |
            SUCCESS | return FALSE | return FALSE    |
                    |              |                 |
                    +--------------+-----------------+
                    |              |                 |
            FAILURE | return FALSE | return FALSE    |
                    |              |                 |
                    +--------------+-----------------+
             */
            // login was no succeed
            if (isCommitSucceeded()) {
                // this typical never happens (but who knows the subclasses...)
                return logout();
            } else {
                return clearInternalState();
            }
        }
    }

    /**
     * Method commit checks if the login succeeded and if true it calls {@link #commitState()} to
     * apply all principals and credentials to the subject.
     *
     * @return boolean true if commit succeeded.
     * @throws LoginException if error occurs.
     * @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jaas/JAASLMDevGuide.html#commit">
     *      Description of the commit method in the LoginModule Developer's Guide
     *      </a>
     */
    public boolean commit() throws LoginException {
        if (LOG.isDebugEnabled()) { LOG.debug("AbstractLoginModule.commit"); }
        /*
        LOGIN \ COMMIT
                \   SUCCESS          FAILURE
                +--------------+-----------------+
                |              |                 |
        SUCCESS | return TRUE  | throw EXCEPTION |
                |              |                 |
                +--------------+-----------------+
                |              |                 |
        FAILURE | return FALSE | return FALSE    |
                |              |                 |
                +--------------+-----------------+
        */
        if (isLoginSucceeded()) {
            // only if previous login was a success commit the internal state
            return commitState();
        } else {
            // otherwise return false and clear the user credentials.
            clearAuthenticationState();
            return false;
        }
    }

    /**
     * Method logout releases all state information applied during a commit from the subject.
     *
     * @return boolean.
     * @throws LoginException if error occurs.
     * @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jaas/JAASLMDevGuide.html#logout">
     *      Description of the logout method in the LoginModule Developer's Guide
     *      </a>
     */
    public boolean logout() throws LoginException {
        if (LOG.isDebugEnabled()) { LOG.debug("AbstractLoginModule.logout"); }
        if (isCommitSucceeded()) {
            // if principals and public/private credentials were applied to the subject
            // clean all principals and credentials from the subject
            Subject subject = getSubject();
            if (!subject.isReadOnly()) {
                subject.getPrincipals().removeAll(getPrincipals());
                subject.getPrivateCredentials().removeAll(getPrivateCredentials());
                subject.getPublicCredentials().removeAll(getPublicCredentials());
                // destroy credentials silent it is not necessary to throw an exception
                // because all credentails will be garbaged.
                destroy(getPublicCredentials());
                destroy(getPrivateCredentials());
            } else {
                // the subject is in readonly state,therefor only the credentials can possible be destroyed.
                if (LOG.isWarnEnabled()) {
                    LOG.warn(
                            "subject is readonly, commited principals can't be removed, try to destroy credentials.");
                }
                boolean allpubliccredentialsdestroyed = destroy(getPublicCredentials());
                boolean allprivatecredentialsdestroyed = destroy(getPrivateCredentials());
                if (!allpubliccredentialsdestroyed) {
                    throw new LoginException(
                            "subject is readonly, not all public credentials could be destroyed during logout.");
                }
                if (!allprivatecredentialsdestroyed) {
                    throw new LoginException(
                            "subject is readonly, not all private credentials could be destroyed during logout.");
                }
            }
        }
        // now reset the internal state
        clearInternalState();
        return true;
    }

    /**
     * Method checks the some default settings before it delegates to the performLogin method.
     * It checks if the loginmodule is disabled and that a call backhandler is provided.
     *
     * @return boolean.
     * @throws LoginException if error occurs.
     */
    public boolean login() throws LoginException {
        if (LOG.isDebugEnabled()) { LOG.debug("AbstractLoginModule.login"); }
        Options options = getOptions();
        if (options.isDisabled()) {
            if (LOG.isInfoEnabled()) {
                LOG.info("login module is disabled -> login module will be ignored.");
            }
            return false;
        }
        if (!isCallbackHandlerSet()) {
            // login modules need a callback handler to retrieve the username and password.
            // if no callback handler is set, the login module thought that it shall be ignored.
            if (LOG.isWarnEnabled()) {
                LOG.warn("no callback handler set -> login module will be ignored.");
            }
            return false;
        }
        if (options.isModuleBanner()) {
            showModuleBanner();
        }
        if (performLogin()) {
            setLoginSucceeded(true);
            return true;
        }
        return false;
    }

    /**
     * Method performs the real login, this is called by the pre implented login() method.
     *
     * @return boolean.
     * @throws LoginException if error occurs.
     * @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jaas/JAASLMDevGuide.html#login">
     *      Description of the login method in the LoginModule Developer's Guide
     *      </a>
     */
    protected abstract boolean performLogin() throws LoginException;

    /**
     * Method initialize the login module.
     * <p/>
     * All parameters are stored in internal variables, accessible with getter methods.
     * The sharedState and options parameter maps are wrapped with special classes
     * {@link org.n52.security.authentication.loginmodule.SharedState}
     * and {@link org.n52.security.authentication.loginmodule.Options}.
     * <p/>
     *
     * @param subject         the subject.
     * @param callbackHandler the callbackHandler can be null.
     * @param sharedState     the sharedState map.
     * @param options         the options map.
     * @see javax.security.auth.spi.LoginModule#initialize(javax.security.auth.Subject,javax.security.auth.callback.CallbackHandler,java.util.Map,java.util.Map)
     * @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jaas/JAASLMDevGuide.html#initialize">
     *      Description of the initialize method in the LoginModule Developer's Guide
     *      </a>
     */
    public void initialize(final Subject subject,
                           final CallbackHandler callbackHandler,
                           final Map sharedState,
                           final Map options) {
        if (LOG.isDebugEnabled()) { LOG.debug("AbstractLoginModule.initialize"); }
        assert subject != null;
        assert sharedState != null;
        assert options != null;

        m_subject = subject;
        m_callbackHandler = callbackHandler;
        m_sharedState = new SharedState(sharedState);
        m_options = new Options(options);
        m_principals = new HashSet();
        m_publicCredentials = new HashSet();
        m_privateCredentials = new HashSet();
        // initialize sub class
        initialize();
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Method adds a principal to the local principal set.
     * The principals will be added to the subject in the commit method and removed
     * from the subject on logout or abort.
     *
     * @param principal a principal created during login.
     */
    protected void addPrincipal(final Principal principal) {
        m_principals.add(principal);
    }

    /**
     * Method adds a private credential to the local private credential set.
     * The credentials will be added to the subject in the commit method and removed
     * from the subject on logout or abort.
     *
     * @param credential a private credential created during login.
     */
    protected void addPrivateCredential(final Object credential) {
        m_privateCredentials.add(credential);
    }

    /**
     * Method adds a public credential to the local public credential set.
     * The credentials will be added to the subject in the commit method and removed
     * from the subject on logout or abort.
     *
     * @param credential a public credential created during login.
     */
    protected void addPublicCredential(final Object credential) {
        m_publicCredentials.add(credential);
    }

    /**
     * Method clearInternalState clears all internal principals and credentials and sets the
     * state of loginSucceeded and commitSucceeded back to false and clears all references to
     * external resouces (e.g. CallbackHandler). It also invokes the {@link #clearAuthenticationState()}
     * method to ensure that no user credentials are further stored.
     *
     * @return boolean.
     * @throws LoginException if error occurs.
     */
    protected boolean clearInternalState() throws LoginException {
        // remove principals and credentials
        getPrincipals().clear();
        getPrivateCredentials().clear();
        getPublicCredentials().clear();
        // reset internal login state
        setCommitSucceeded(false);
        setLoginSucceeded(false);

        // remove possible stored user authentication credentials (e.g. password)
        clearAuthenticationState();

        // clear external resources
        m_subject = null;
        m_callbackHandler = null;
        m_options = null;
        m_sharedState = null;
        return true;
    }

    /**
     * Method cleanAuthenticationState is a template method used to invoke a cleanup of
     * any authentication information provided by the user e.g. the user name and password used
     * to login the user.
     * <p/>
     * This is not a clean up of the created principals or credentials during login.
     * It is a way to reduce the security holes by cleaning state information wich is not further needed.
     *
     * @throws javax.security.auth.login.LoginException
     *          if an error occurs.
     */
    protected abstract void clearAuthenticationState() throws LoginException;

    /**
     * Method commitState applies all principals, public and private credentials
     * created during the login or the prepareCommitState method to the subject
     * and sets the internal commitSucceded state to true.
     * It also calls {@link #clearAuthenticationState()}.
     *
     * @return boolean.
     * @throws javax.security.auth.login.LoginException
     *          if commit fails e.g. because the subject is readonly.
     * @see #login()
     * @see #addPrincipal(java.security.Principal)
     * @see #addPrivateCredential(Object)
     * @see #addPublicCredential(Object)
     */
    protected boolean commitState() throws LoginException {
        Subject subject = getSubject();
        if (subject.isReadOnly()) {
            throw new LoginException(
                    "subject is readonly neither principals nor credentials can be added.");
        }
        if (LOG.isDebugEnabled()) {
            LOG.debug("login succeeded, now try to commit the principals.");
        }
        // let sub class prepare some principals and credentials for commit
        prepareCommitState();
        // now apply all informations.
        subject.getPrincipals().addAll(getPrincipals());
        subject.getPrivateCredentials().addAll(getPrivateCredentials());
        subject.getPublicCredentials().addAll(getPublicCredentials());
        setCommitSucceeded(true);
        clearAuthenticationState();
        return true;
    }

    /**
     * Method prepareCommitState is a template method used to let a sub class prepare all
     * principals and credentials, which then will be applied to the subject.
     * <p/>
     * It allows to separate the logic of the login (retrieve and check the authentication credentials)
     * from the logic for retrieving/preparing subject information.
     *
     * @throws LoginException if error occurs.
     */
    protected abstract void prepareCommitState() throws LoginException;

    /**
     * Method destroy checks the set for instances of {@link javax.security.auth.Destroyable} and destroys them.
     * <p/>
     * The method only returns true if all instances within the set are Destroyables
     * and the destroy method succeeded on all.
     *
     * @param destroyables the destroyables.
     * @return boolean flag indicates that all instances are destroyed.
     */
    protected boolean destroy(final Set destroyables) {
        boolean alldestroyed = true;
        for (Iterator iterator = destroyables.iterator(); iterator.hasNext();) {
            Object o = iterator.next();
            if (o instanceof Destroyable) {
                try {
                    ((Destroyable) o).destroy();
                } catch (DestroyFailedException e) {
                    if (LOG.isWarnEnabled()) {
                        LOG.warn("principal or credentail could not be destroyed!", e);
                    }
                    alldestroyed = false;
                }
            } else {
                alldestroyed = false;
            }
        }
        return alldestroyed;
    }

    /**
     * Method initialize replaces the
     * {@link javax.security.auth.spi.LoginModule#initialize(javax.security.auth.Subject,javax.security.auth.callback.CallbackHandler,java.util.Map,java.util.Map)}.
     * <p/>
     * All parameters of the original initialize method can be retrieved by getter methods.
     * <p/>
     * A subclasses shall check the options and initializes them self.
     *
     * @see javax.security.auth.spi.LoginModule#initialize(javax.security.auth.Subject,javax.security.auth.callback.CallbackHandler,java.util.Map,java.util.Map)
     * @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jaas/JAASLMDevGuide.html#initialize">
     *      Description of the initialize method in the LoginModule Developer's Guide
     *      </a>
     */
    protected abstract void initialize();

    /**
     * Method gets the callbackHandlerSet of this AbstractLoginModule object.
     *
     * @return boolean.
     */
    protected boolean isCallbackHandlerSet() {
        return m_callbackHandler != null;
    }

    /**
     * Method showModuleBanner sends a description of the login module to the callback handler.
     * This method shall be used in login methods if the option
     * {@link Options#isModuleBanner()} is set.
     *
     * @throws LoginException if a error occurs within the callback handler during this operation.
     * @see org.n52.security.authentication.loginmodule.Options#OPTION_MODULEBANNER
     */
    protected void showModuleBanner() throws LoginException {
        // send description of this login module to the callbackhandler
        Callback[] callbacks = new Callback[]{
                new TextOutputCallback(TextOutputCallback.INFORMATION, getDescription())
        };
        handleCallbacks(callbacks);
    }

    /**
     * Method gets the description of this LoginModule.
     * <p/>
     * The description is used if the login module is configured with the option 'moduleBanner',
     * to invoke a {@link javax.security.auth.callback.TextOutputCallback} on the callback handler.
     *
     * @return String.
     * @see org.n52.security.authentication.loginmodule.Options#OPTION_MODULEBANNER
     * @see #showModuleBanner()
     */
    protected abstract String getDescription();

    /**
     * Method handleCallbacks sends the given callback array to the callback handler.
     * <p/>
     * It converts a eventually thrown IOException or UnsuportedCallbackException
     * to a LoginException, therefore use this method only if there is no other possible reaction to
     * the exceptions thrown by the CallbackHandler.
     *
     * @param callbacks the callbacks to execute.
     * @throws LoginException if error occurs in the callback handler.
     */
    protected void handleCallbacks(final Callback[] callbacks) throws LoginException {
        try {
            CallbackHandler handler = getCallbackHandler();
            handler.handle(callbacks);
        } catch (IOException e) {
            throw new LoginException(
                    "[Error] a io error occured in the callback handler. Cause:" + e);
        } catch (UnsupportedCallbackException e) {
            throw new LoginException("[Error] " + e.getCallback().toString()
                                     + " not available to garner authentication information "
                                     + "from the user. Cause:" + e);
        }
    }
}
