/*******************************************************************************
 * Copyright  2007 52North Initiative for Geospatial Open Source Software GmbH
 *
 * Author: Marko Reiprecht, con terra GmbH
 *
 * Contact: Andreas Wytzisk, 52North Initiative for Geospatial Open Source
 * Software GmbH, Martin-Luther-King-Weg 24, 48155 Muenster, Germany,
 * info@52north.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; even without the implied WARRANTY OF MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundations web page, http://www.fsf.org.
 *
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.authentication.loginmodule;

import java.io.Serializable;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.login.LoginException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Class AbstractPasswordLoginModule provides a frame implementation for login modules which use
 * username and password credentials to authenticate a user.
 * <p/>
 * This class complete supports the PAM flags try_first_pass, try_mapped_pass, use_first_pass and
 * use_mapped_pass in invokes according to their rules the {@link #login(String,char[])} method.
 * <p/>
 * It also supports the
 * {@link org.n52.security.authentication.loginmodule.Options#OPTION_MODULEBANNER moduleBanner} and
 * {@link org.n52.security.authentication.loginmodule.Options#OPTION_DISABLED disabled} flag.
 * <p/>
 * Sub classes have to implement the methods:<br/>
 * {@link #getDescription()}<br/>
 * {@link #initialize()}<br/>
 * {@link #login(String,char[])} and<br/>
 * {@link #prepareCommitState()}.<br/>
 *
 * @author Marko Reiprecht
 *         Created on 16.09.2007
 * @version $Revision: $
 * @see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/security/jaas/JAASLMDevGuide.html">
 *      the LoginModule Developer's Guide for more information about the PAM rules
 *      </a>
 */
public abstract class AbstractPasswordLoginModule extends AbstractLoginModule
        implements Serializable {

    // ------------------------------ FIELDS ------------------------------

    /**
     * The Logger.
     */
    private static final Log LOG = LogFactory.getLog(AbstractPasswordLoginModule.class);
    /**
     * only a empty char used to reset the password.
     */
    private static final char[] EMPTY_CHAR_ARRAY = new char[0];
    /**
     * The serialVersionUID.
     *
     * @see java.io.Serializable
     */
    private static final long serialVersionUID = -3891187066544947907L;

    /**
     * Field m_username holds the username gained from a callback handler or the sharedState.
     */
    private String m_username;
    /**
     * Field m_password holds the password gained from a callback handler or the sharedState.
     */
    private char[] m_password;

    // --------------------- GETTER / SETTER METHODS ---------------------

    /**
     * Method gets the password.
     *
     * @return char[].
     */
    protected char[] getPassword() {
        return m_password;
    }

    /**
     * Method gets the username.
     *
     * @return String.
     */
    protected String getUsername() {
        return m_username;
    }

    // ------------------------ INTERFACE METHODS ------------------------


    // --------------------- Interface LoginModule ---------------------

    /**
     * Method performs login via username/password after the PAM-Rules.
     * <p/>
     * This means that the flags try_first_pass, try_mapped_pass, use_first_pass and
     * use_mapped_pass are recognized. If one of these are true, the login module will check the
     * shared state for already provided credentials and use them for the login.
     * <p/>
     * It also checks for the moduleBanner option and invokes a TextOutputCallback if the flag is true.
     *
     * @return boolean.
     * @throws LoginException if error occurs.
     * @see org.n52.security.authentication.loginmodule.Options#OPTION_TRY_FIRST_PASS
     * @see org.n52.security.authentication.loginmodule.Options#OPTION_TRY_MAPPED_PASS
     * @see org.n52.security.authentication.loginmodule.Options#OPTION_USE_FIRST_PASS
     * @see org.n52.security.authentication.loginmodule.Options#OPTION_USE_MAPPED_PASS
     */
    public boolean performLogin() throws LoginException {
        if (LOG.isDebugEnabled()) { LOG.debug("AbstractPasswordLoginModule.performLogin"); }
        Options options = getOptions();
        SharedState sharedState = getSharedState();
        // check for special PAM-Options (reuse of allready known username and password)
        if (sharedState.isLoginNameSet() && sharedState.isLoginPasswordSet()
            && (options.isTryFirstPass()
                || options.isUseFirstPass()
                || options.isTryMappedPass()
                || options.isUseMappedPass())) {
            m_username = sharedState.getLoginName();
            m_password = sharedState.getLoginPassword();

            // try to login with shared credentials if it fails this module asks for new credentials
            if (options.isTryFirstPass() || options.isTryMappedPass()) {
                try {
                    return performLogin(m_username, m_password);
                } catch (LoginException e) {
                    // ignore exception and try to retrieve new credentials for this login module
                    if (LOG.isInfoEnabled()) {
                        LOG.info(
                                "first try to login with credentials from the previous login module failed",
                                e);
                    }
                    m_username = null;
                    m_password = null;
                }
            }
        }
        if (m_username == null || m_password == null) {
            // get username and password from the callback handler
            Callback[] callbacks = new Callback[2];
            callbacks[0] = new NameCallback("user name: ");
            callbacks[1] = new PasswordCallback("password: ", false);
            handleCallbacks(callbacks);
            m_username = ((NameCallback) callbacks[0]).getName();
            m_password = ((PasswordCallback) callbacks[1]).getPassword();
            if (m_username == null && m_password == null) {
                // seems that no credentials for this module are present, so this module goes to do nothing
                // may be other modules are able to authenticate the user.
                if (LOG.isInfoEnabled()) {
                    LOG.info("No credentials for module <" + getClass().getName() + "> available, skip login.");
                }
                return false;
            }
            if (m_username == null) {
                // null == empty username
                m_username = "";
            }
            if (m_password == null) {
                // null == empty password
                m_password = EMPTY_CHAR_ARRAY;
            }
        }
        // now try login without catching any LoginException
        return performLogin(m_username, m_password);
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * This clears the internaly stored username and password.
     *
     * @throws LoginException if error occurs.
     * @see AbstractLoginModule#clearAuthenticationState()
     */
    protected void clearAuthenticationState() throws LoginException {
        m_username = null;
        clear(m_password);
        m_password = null;
    }

    /**
     * Method clear replaces all chars in the internal password with empty chars.
     * This only done for security reasons.
     *
     * @param password the password.
     */
    private void clear(final char[] password) {
        if (password == null) {
            return;
        }
        //clear password chars
        for (int i = 0, j = password.length; i < j; ++i) {
            password[i] = ' ';
        }
    }

    /**
     * Method performLogin invokes the {@link #login(String,char[])} method.
     * <p/>
     * If the login method returns true the loginSucceeded state is set to true.
     * <p/>
     * It also checks the options for one of the try_first_pass, try_mapped_pass,
     * use_first_pass, use_mapped_pass flags. If on of these are set
     * and the sharedState contains no username and password than the internal username and
     * password will be stored in the shared state.
     *
     * @param username the username.
     * @param password the password.
     * @return boolean.
     * @throws LoginException if error occurs.
     */
    private boolean performLogin(final String username, final char[] password)
            throws LoginException {
        if (LOG.isDebugEnabled()) { LOG.debug("AbstractPasswordLoginModule.performLogin("+username+")"); }
        if (login(username, password)) {
            SharedState sharedState = getSharedState();
            Options options = getOptions();
            if (!sharedState.isLoginNameSet() && !sharedState.isLoginPasswordSet()
                && (options.isTryFirstPass()
                    || options.isUseFirstPass()
                    || options.isTryMappedPass()
                    || options.isUseMappedPass())) {
                // store password in shared state
                // for use in the next login modules
                sharedState.setLoginName(username);
                sharedState.setLoginPassword(password);
            }
            return true;
        }
        return false;
    }

    /**
     * Method isEqual tests the both char[] on equality.
     *
     * @param allowedPassword the allowedPassword.
     * @param password        the password to check.
     * @param ignorecase      ignores case during check.
     * @return only true if both char[] equal.
     */
    protected boolean isEqual(final char[] allowedPassword, final char[] password,
                              final boolean ignorecase) {
        if (allowedPassword == password) {
            return true;
        }
        if (allowedPassword == null || password == null) {
            return false;
        }
        if (allowedPassword.length != password.length) {
            return false;
        }
        for (int i = 0, j = allowedPassword.length; i < j; ++i) {
            char a = allowedPassword[i];
            char b = password[i];
            if (ignorecase) {
                a = Character.toUpperCase(a);
                b = Character.toUpperCase(b);
            }
            if (a != b) {
                return false;
            }
        }
        return true;
    }

    /**
     * Method login authenticates a user with a given username and password.
     * <p/>
     * An implementation class shall not store the given parameters in member variables,
     * they are allready stored in this class and can be retrieved with
     * the {@link #getUsername()} and {@link #getPassword()} methods.
     *
     * @param username the username.
     * @param password the password.
     * @return boolean.
     * @throws LoginException if error occurs.
     */
    protected abstract boolean login(final String username, final char[] password)
            throws LoginException;
}
