/*******************************************************************************
 * SDI Security Services
 *
 * Copyright (C) 2005 by con terra GmbH
 *
 * Author: Bastian Schaeffer
 *
 * Contact: Albert Remke, con terra GmbH, Martin-Luther-King-Weg 24, 48155
 * Muenster, Germany, 52n@conterra.de
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the web page of the Free Software Foundation, http://www.fsf.org.
 ******************************************************************************/

package org.n52.security.authentication.loginmodule;

import java.io.IOException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;
import org.apache.log4j.Logger;
import org.n52.security.authentication.UsernamePasswordCredential;
import org.n52.security.authentication.callbacks.CredentialCallback;
import org.n52.security.authentication.principals.LoginNamePrincipal;
import org.n52.security.authentication.principals.RolePrincipal;
import org.n52.security.authentication.principals.UsernameIDPrincipal;
import org.n52.security.common.util.StringUtils;

public class DataBaseLoginModule implements LoginModule {

    private Subject subject;

    private CallbackHandler callbackHandler;

    private Map sharedState;

    private Map options;

    private boolean debug = false;

    private Set principals;
    private static Connection sDBConnection = null;

    private static Logger sLogger = Logger.getLogger(DataBaseLoginModule.class);

    /*
     * (non-Javadoc)
     *
     * @see javax.security.auth.spi.LoginModule#abort()
     */
    public boolean abort() throws LoginException {
        subject = null;
        callbackHandler = null;
        sharedState = null;
        options = null;
        debug = false;
        principals.clear();

        return true;
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.security.auth.spi.LoginModule#commit()
     */
    public boolean commit() throws LoginException {
        subject.getPrincipals().addAll(principals);
        return true;
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.security.auth.spi.LoginModule#login()
     */
    public boolean login() throws LoginException {

        Statement stmt = null;
        ResultSet rs = null;

        boolean isBase64Encoded = Boolean.valueOf(
                (String) options.get("credential.isBase64Encoded")).booleanValue();

        try {

            // get the initialization params

            if (sDBConnection == null) {
                String dbDriverClass = (String) options.get("db.driver.class");
                String dbURL = (String) options.get("db.url");
                String dbUser = (String) options.get("db.user");
                String dbPassword = (String) options.get("db.password");
                sLogger.info("Creating database connection " + dbURL + ", " + dbUser + ", " + dbPassword);
                Class.forName(dbDriverClass);
                sDBConnection = DriverManager.getConnection(dbURL, dbUser, dbPassword);
            }

            // get the credentials
            Set crendentialSet = subject.getPublicCredentials();
            Iterator it = crendentialSet.iterator();
            UsernamePasswordCredential credential = null;
            while (it.hasNext()) {
                Object publicCred = it.next();
                if (publicCred instanceof UsernamePasswordCredential) {
                    credential = (UsernamePasswordCredential) it.next();
                    break;
                }
            }
            if (credential == null) {
                // use callbackhandler
                CredentialCallback credentialCallback = new CredentialCallback(UsernamePasswordCredential.class);
                callbackHandler.handle(new Callback[]{credentialCallback});
                credential = (UsernamePasswordCredential) credentialCallback.getCredential();
                if (credential == null) {
                    throw new LoginException("no required credentials found");
                }
            }

            // check if credentials are given properly

            String user = credential.getUsername();
            String pass = new String(credential.getPassword());

            if (isBase64Encoded) {
                pass = StringUtils.decodeBase64(pass);
                user = StringUtils.decodeBase64(user);

            }
            sLogger.debug("Trying to authenticate " + user);

            stmt = sDBConnection.createStatement();
            String statement = "SELECT Name, EMail, Password FROM Users WHERE Username = '" + user
                               + "';";
            rs = stmt.executeQuery(statement);

            boolean tUserExists = rs.next();
            if (!tUserExists) {
                throw new LoginException("Unknown login name");
            }

            if (!rs.getString("Password").equals(pass)) {
                throw new LoginException("Wrong password");
            }

            principals.add(new LoginNamePrincipal(user));
            principals.add(new UsernameIDPrincipal(user));
            principals.add(new RolePrincipal(rs.getString("Name")));
            return true;

        }
        catch (ClassNotFoundException cnfe) {
            sLogger.error("Database driver class not found", cnfe);
            throw new LoginException("Internal service error. Please contact the service administrator.");
        }
        catch (SQLException se) {
            sLogger.error("Server-side database _access_ error: " + se.getMessage(), se);
            throw new LoginException("Internal service error. Please contact the service administrator.");

        }
        catch (UnsupportedCallbackException e) {
            throw new LoginException("CallbackHandler supports no CredentialCallbacks.");
        } catch (IOException e) {
            throw new LoginException("Internal service error. Please contact the service administrator.");
        } finally {
            try {
                rs.close();
                stmt.close();
            }
            catch (Exception e) {
                sLogger.error("Error during login process", e);
                throw new LoginException("Internal service error. Please contact the service administrator.");

            }
        }
    }

    /*
     * (non-Javadoc) /* (non-Javadoc)
     *
     * @see javax.security.auth.spi.LoginModule#logout()
     */
    public boolean logout() throws LoginException {
        Iterator it = subject.getPrincipals().iterator();

        while (it.hasNext()) {
            subject.getPrincipals().remove(it.next());
        }
        return true;
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.security.auth.spi.LoginModule#initialize(javax.security.auth.Subject,
     *      javax.security.auth.callback.CallbackHandler, java.util.Map, java.util.Map)
     */
    public void initialize(Subject subject,
                           CallbackHandler callbackHandler,
                           Map sharedState,
                           Map options) {
        this.subject = subject;
        this.callbackHandler = callbackHandler;
        this.sharedState = sharedState;
        this.options = options;
        debug = "true".equalsIgnoreCase((String) options.get("debug"));
        this.principals = new HashSet(8);
    }
}
