/*******************************************************************************
 * Copyright  2007 52North Initiative for Geospatial Open Source Software GmbH
 *
 * Author: Marko Reiprecht, con terra GmbH
 *
 * Contact: Andreas Wytzisk, 52North Initiative for Geospatial Open Source
 * Software GmbH, Martin-Luther-King-Weg 24, 48155 Muenster, Germany,
 * info@52north.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; even without the implied WARRANTY OF MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundations web page, http://www.fsf.org.
 *
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.authentication.loginmodule;

import java.io.Serializable;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.n52.security.authentication.principals.LoginNamePrincipal;
import org.n52.security.authentication.principals.RolePrincipal;
import org.n52.security.authentication.principals.UsernameIDPrincipal;
import org.n52.security.common.crypto.DigestUtil;

/**
 * Class SingleUserLoginModule is a simple login module wich only knows a single user.
 * <p/>
 * This user and his credentials are configured by the options of this login module.<br/>
 * {@link #OPTION_ALLOWED_USERNAME allowedUsername} defines the name of the allowed user.<br/>
 * {@link #OPTION_ALLOWED_PASSWORD allowedPassword} defines the password of the allowed user.<br/>
 * {@link #OPTION_PASSWORD_ENCRYPTION_ALG pwdEncAlg} defines the encryption allgorithmen wihtin the password is stored,
 * valid values are 'plain','md5', 'sha'.<br/>
 * {@link #OPTION_USERROLES userRoles} defines a pipe separated list of user roles, wich will be applied as
 * {@link org.n52.security.authentication.principals.RolePrincipal}s if the login succeeded.
 * <p/>
 * Besides the role principals, creates this login module also
 * a {@link org.n52.security.authentication.principals.LoginNamePrincipal}
 * which holds the login name identiy of the user and his authentication date.
 *
 * @author Marko Reiprecht
 *         Created on 16.09.2007
 * @version $Revision: $
 */
public class SingleUserLoginModule extends AbstractPasswordLoginModule implements Serializable {

    // ------------------------------ FIELDS ------------------------------

    /**
     * Field OPTION_ALLOWED_PASSWORD identifies the 'allowedPassword' option.
     * A m_value of this option must be a valid password (encrypted or plain)
     *
     * @see #OPTION_PASSWORD_ENCRYPTION_ALG
     */
    protected static final String OPTION_ALLOWED_PASSWORD = "allowedPassword";
    /**
     * Field OPTION_PASSWORD_ENCRYPTION_ALG  identifies the 'pwdEncAlg' option.
     * A m_value of this option must be a valid encryption type.
     * Valid encryption algorithms are:<br/>
     * 'plain'  - no encryption
     * 'md5'    - MD5 Hash
     * 'sha'   - SHA Hash
     */
    protected static final String OPTION_PASSWORD_ENCRYPTION_ALG = "pwdEncAlg";
    /**
     * Field OPTION_ALLOWED_USERNAME  identifies the 'allowedUsername' option.
     * A m_value of this option must be a valid username.
     */
    protected static final String OPTION_ALLOWED_USERNAME = "allowedUsername";

    /**
     * Field OPTION_USERROLES  identifies the 'userRoles' option.
     * A m_value of this option must be a pipe separated string of role names.
     */
    protected static final String OPTION_USERROLES = "userRoles";

    /**
     * The Logger.
     */
    private static final Log LOG = LogFactory.getLog(SingleUserLoginModule.class);
    /**
     * Field PW_ENC_ALG_PLAIN  holds the default m_value for "no-encryption" of the password.
     */
    private static final String PW_ENC_ALG_PLAIN = "plain";
    /**
     * The serialVersionUID.
     *
     * @see java.io.Serializable
     */
    private static final long serialVersionUID = -97291032500611905L;
    /**
     * Field EMPTY_STRING_ARRAY is only a empty string to preserver allocation.
     */
    private static final String[] EMPTY_STRING_ARRAY = new String[0];

    /**
     * Field m_allowedUserName holds the allowed username.
     */
    private String m_allowedUserName;
    /**
     * Field m_allowedPassword holds the allowed password.
     */
    private char[] m_allowedPassword;

    /**
     * Field m_passwordEncryptionAlg holds the password encryption algorithm.
     */
    private String m_passwordEncryptionAlg;

    /**
     * Field m_userRoles holds a list of role names.
     */
    private String[] m_userRoles;

    /**
     * Field m_plainPW  indicats that the password is not encrypted.
     */
    private boolean m_plainPW;

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Gets a short description of this login module.
     *
     * @return String message.
     * @see AbstractLoginModule#getDescription()
     */
    protected String getDescription() {
        return new StringBuffer(150)
                .append("The authentication is performed through the '")
                .append(SingleUserLoginModule.class.getName())
                .append("' login module. It requires a user name and password.").toString();
    }

    /**
     * This initializes the login module.
     * <p/>
     * It gets the values for the options {@link #OPTION_ALLOWED_USERNAME allowedUsername} and
     * {@link #OPTION_ALLOWED_PASSWORD allowedPassword} from the login module options and stores
     * them localy.
     *
     * @see AbstractLoginModule#initialize()
     */
    protected void initialize() {
        Options options = getOptions();

        // get username, default is empty
        m_allowedUserName = options.getAsString(OPTION_ALLOWED_USERNAME, "");

        // get password, default is empty
        m_allowedPassword = options.getAsString(OPTION_ALLOWED_PASSWORD, "").toCharArray();

        // gets the encryption type of the password, default is plain(no encryption)
        m_passwordEncryptionAlg = options.getAsString(OPTION_PASSWORD_ENCRYPTION_ALG,
                                                      PW_ENC_ALG_PLAIN);
        // store if password is plain.
        m_plainPW = PW_ENC_ALG_PLAIN.equalsIgnoreCase(m_passwordEncryptionAlg);
        // test if encryption algorithm is supported wihthin the VM.
        if (!m_plainPW) {
            try {
                if (!DigestUtil.isDigestOf(m_allowedPassword, m_passwordEncryptionAlg)) {
                    if (LOG.isErrorEnabled()) {
                        String msg = new StringBuffer().append(
                                "password has not the correct format for algorithm:").append(
                                m_passwordEncryptionAlg).toString();
                        LOG.error(msg);
                        throw new IllegalArgumentException(msg);
                    }
                }
            } catch (NoSuchAlgorithmException e) {
                if (LOG.isErrorEnabled()) {
                    LOG.error("password encryption algorithm '" + m_passwordEncryptionAlg
                              + "' is not supported, or no MessageDigest.", e);
                }
            }
        }

        // gets the roles for the user, default is no role.
        m_userRoles = options.getAsStringArray(OPTION_USERROLES, EMPTY_STRING_ARRAY, "|");


        if (m_allowedUserName.length() == 0 && LOG.isWarnEnabled()) {
            LOG.warn("option '" + OPTION_ALLOWED_USERNAME
                     + "' is empty, this can be a security risk!");
        }
        if (m_allowedPassword.length == 0 && LOG.isWarnEnabled()) {
            LOG.warn("option '" + OPTION_ALLOWED_PASSWORD
                     + "' is empty, this can be a security risk!");
        }
        if (m_plainPW && LOG.isWarnEnabled()) {
            LOG.warn("option '" + OPTION_PASSWORD_ENCRYPTION_ALG
                     + "' contains a value of '" + PW_ENC_ALG_PLAIN + "', this can be a security risk!");
        }
        if (m_userRoles.length == 0 && LOG.isWarnEnabled()) {
            LOG.warn("option '" + OPTION_USERROLES
                     + "' is empty, therewith authenticated users will have no role prinicipals.");
        }
    }

    /**
     * This checks for the correct username and password, they must be equal
     * to the values of the options {@link #OPTION_ALLOWED_USERNAME allowedUsername} and
     * {@link #OPTION_ALLOWED_PASSWORD allowedPassword}.
     * <p/>
     * If the options "try_mapped_pass" or "use_mapped_pass" are set and the allowed password is
     * encrypted, than the method tries to calculate the hash for the given password.
     * Otherwise the password is treaten as string representation of the hash and directly compared.
     * A string representation of one hash byte is a his hexadecimal m_value written with two digits.
     *
     * @param username the username.
     * @param password the password.
     * @return true if login succeeded, otherwise an exception is thrown.
     * @throws LoginException if the username or password are incorrect
     *                        or the login module is wrong configured.
     * @see AbstractPasswordLoginModule#login(String,char[])
     */
    protected boolean login(final String username, final char[] password)
            throws LoginException {
        if (LOG.isTraceEnabled()) { LOG.trace("SingleUserLoginModule.login("+username+")"); }
        // check username
        if (m_allowedUserName.equals(username)) {
            // check for plain passwords
            if (m_plainPW && isEqual(m_allowedPassword, password, false)) {
                return true;
            } else if (!m_plainPW) {
                // password is message digest:
                Options options = getOptions();
                if (options.isTryMappedPass() || options.isUseMappedPass()) {
                    // if "try_mapped_pass" or "use_mapped_pass" is set
                    // then the password is entered plain, therewith the digest must be calculated.
                    byte[] pwhash = getMessageDigest().digest(
                            Charset.forName("UTF-8").encode(CharBuffer.wrap(password)).array());
                    // now convert hash back to chars and compare with stored pw
                    char[] pwhashchars = DigestUtil.digestToCharArray(pwhash);
                    if (isEqual(m_allowedPassword, pwhashchars, true)) {
                        return true;
                    }
                } else {
                    // otherwise both values are the hashes in hex-digits.
                    // so only compare the values, and the hash format.
                    if (isEqual(m_allowedPassword, password, true)
                        // this is only for internal stability,
                        // for the case that someone change the behavior of the initialize method
                        && DigestUtil.isDigestOf(password, getMessageDigest())) {
                        return true;
                    }
                }
            }
        }
        throw new FailedLoginException("wrong username or password.");
    }

    /**
     * Method gets the messageDigest creates a MessageDigest for
     * the {@link #OPTION_PASSWORD_ENCRYPTION_ALG password encryption algorithm}.
     *
     * @return MessageDigest.
     * @throws LoginException MessageDigest can not be created.
     */
    private MessageDigest getMessageDigest() throws LoginException {
        try {
            return MessageDigest.getInstance(m_passwordEncryptionAlg);
        } catch (NoSuchAlgorithmException e) {
            throw new LoginException(
                    new StringBuffer().append(
                            "login module wrong configured. Check the '").append(
                            OPTION_PASSWORD_ENCRYPTION_ALG).append(
                            "' for a correct m_value. Cause: ").append(e).toString());
        }
    }

    /**
     * Prepares a LoginNamePrincipal with out of the username,
     * and some RolePrincipals from the userRoles option,
     * all within the {@link org.n52.security.authentication.principals.Scope#GLOBAL global scope}.
     *
     * @throws LoginException if an error occurs.
     * @see AbstractLoginModule#prepareCommitState()
     */
    protected void prepareCommitState() throws LoginException {
        // carefull check if login is succeeded (thats not realy needed)
        if (isLoginSucceeded()) {
            // add login name principal, the name is in the global scope.
            addPrincipal(new LoginNamePrincipal(getUsername()));
            addPrincipal(new UsernameIDPrincipal(getUsername()));
            // add roles
            for (int i = 0, l = m_userRoles.length; i < l; ++i) {
                String userRole = m_userRoles[i];
                // create RolePrincipals all role names are in the global scope.
                addPrincipal(new RolePrincipal(userRole));
            }
        }
    }
}
