/*******************************************************************************
 * Copyright  2007 52North Initiative for Geospatial Open Source Software GmbH
 *
 * Author: Marko Reiprecht, con terra GmbH
 *
 * Contact: Andreas Wytzisk, 52North Initiative for Geospatial Open Source
 * Software GmbH, Martin-Luther-King-Weg 24, 48155 Muenster, Germany,
 * info@52north.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; even without the implied WARRANTY OF MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundations web page, http://www.fsf.org.
 *
 ******************************************************************************/
/* $Id: AttributeValueAdapter.java 852 2007-12-12 18:59:43Z jessebluemr $ */
package org.n52.security.common.attributes;

import java.io.Serializable;

/**
 * The class AttributeValueAdapter implements all methods of AttributeValue.
 * <p/>
 * It is a helper class for implementing concrete attribute value classes.
 *
 * @author Marko Reiprecht
 * @version $Revision: 852 $
 */
public abstract class AttributeValueAdapter implements AttributeValue, Serializable {

    // ------------------------------ FIELDS ------------------------------

    /**
     * Field serialVersionUID.
     */
    private static final long serialVersionUID = 3478325459333197521L;

    /**
     * The concrete value.
     */
    protected Object m_wrappedObject;

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * creates new AttributeValueAdapter.
     *
     * @param wrappedObject the concrete value object, asserts wrappedObject != null.
     * @throws NullPointerException if wrappedObject is null.
     */
    public AttributeValueAdapter(final Object wrappedObject) throws NullPointerException {
        if (wrappedObject == null) {
            throw new NullPointerException("wrappedObject must not null");
        }
        m_wrappedObject = wrappedObject;
    }

    // ------------------------ CANONICAL METHODS ------------------------

    /**
     * This delegates the equality check to the
     * internal 'wrapped' value.
     *
     * @param o the object to test with.
     * @return true only when o.getUnspecifiedValue().equals(this.getUnspecifiedValue())
     */
    public boolean equals(final Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        final AttributeValueAdapter that = (AttributeValueAdapter) o;

        return m_wrappedObject.equals(that.m_wrappedObject);
    }

    /**
     * Generates the hashCode of this object.
     * It is forwarded to this.getUnspecifiedValue().hashCode().
     *
     * @return the hashcode for this object.
     */
    public int hashCode() {
        return m_wrappedObject.hashCode();
    }

    /**
     * Generates an informal string.
     * This method should only used for logging purposes.
     *
     * @return a string representation.
     */
    public String toString() {
        final StringBuffer sb = new StringBuffer();
        sb.append("org.n52.security.common.attributes.AttributeValueAdapter");
        sb.append("{unspecifiedValue=").append(getUnspecifiedValue());
        sb.append(", typeName='").append(getTypeName()).append('\'');
        sb.append('}');
        return sb.toString();
    }

    /**
     * Gets the concrete value.
     *
     * @return the wrapped concrete value.
     * @see AttributeValue#getUnspecifiedValue()
     */
    public Object getUnspecifiedValue() {
        return m_wrappedObject;
    }

    // ------------------------ INTERFACE METHODS ------------------------


    // --------------------- Interface AttributeValue ---------------------

    /**
     * gets the type name of the concrete value.
     * this is a short cut of getUnspecifiedValue().getClass().getName()
     *
     * @return a type name.
     * @see AttributeValue#getTypeName()
     */
    public String getTypeName() {
        return m_wrappedObject.getClass().getName();
    }
}
