/*******************************************************************************
 * Copyright  2007 52North Initiative for Geospatial Open Source Software GmbH
 *
 * Author: Marko Reiprecht, con terra GmbH
 *
 * Contact: Andreas Wytzisk, 52North Initiative for Geospatial Open Source
 * Software GmbH, Martin-Luther-King-Weg 24, 48155 Muenster, Germany,
 * info@52north.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; even without the implied WARRANTY OF MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundations web page, http://www.fsf.org.
 *
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.common.crypto;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.regex.Pattern;

/**
 * Class DigestUtil contains some convenient utility methods related to
 * {@link java.security.MessageDigest} tasks.
 * <p/>
 *
 * @author Marko Reiprecht
 *         Created on 19.09.2007
 * @version : $
 */
public final class DigestUtil {

    // ------------------------------ FIELDS ------------------------------

    /**
     * Field BYTE_HEX_PATTERN is the regex for bytes as two hex digits.
     */
    private static final Pattern BYTE_HEX_PATTERN = Pattern.compile(
            "^(?:(?:\\d|[abcdefABCDEF]){2})+$"
    );

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Constructor DigestUtil is private to ensure nobody creates instances
     * of this utility class.
     */
    private DigestUtil() {
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Method digestToString converts a byte[] message digest to its two digit hexdecimal
     * representation as string.
     *
     * @param digest the digest !=null.
     * @return String.
     */
    public static String digestToString(final byte[] digest) {
        return new String(digestToCharArray(digest));
    }

    /**
     * Method digestToCharArray converts a byte[] message digest to its two digit hexdecimal
     * representation as chars.
     *
     * @param digest the digest !=null.
     * @return char[].
     */
    public static char[] digestToCharArray(final byte[] digest) {
        char[] digestchars = new char[digest.length << 1];
        for (int i = 0, j = 0, l = digest.length; i < l; ++i, ++j) {
            // convert one byte to 1 or 2 digit hex number
            String bytechars = Integer.toHexString(digest[i] & 0x00FF);
            switch (bytechars.length()) {
                case 1:
                    digestchars[j] = '0';
                    digestchars[++j] = bytechars.charAt(0);
                    break;
                default:
                    digestchars[j] = bytechars.charAt(0);
                    digestchars[++j] = bytechars.charAt(1);
                    break;
            }
        }
        return digestchars;
    }

    /**
     * Method isDigestOf test if the given byte[] array is a digest.
     * This simple tests for the correct length and is only supplied for convenience.
     *
     * @param digest       the digest !=null
     * @param expectedSize the expectedSize.
     * @return boolean.
     */
    public static boolean isDigestOf(final byte[] digest, final int expectedSize) {
        return digest.length == expectedSize;
    }

    /**
     * Method isDigestOf test if the given char[] array is a digest.
     * This tests the correct length and if all chars are in a hexadecimal range.
     * <p/>
     * The expectedSize must be the length of the digest in bytes, not in chars!
     *
     * @param digest       the digest !=null.
     * @param expectedSize the expectedSize of the digest in bytes as given by the algoritm
     *                     {@link java.security.MessageDigest#getDigestLength()}
     *                     Therewith this is the size in bytes, not the size of the char array,
     *                     e.g. MD5 has a size of 16, and the char[] is tested for a length of 32.
     * @return boolean.
     */
    public static boolean isDigestOf(final char[] digest, final int expectedSize) {
        return digest.length == (expectedSize << 1)
               && BYTE_HEX_PATTERN.matcher(new String(digest)).matches();
    }

    /**
     * Method isDigestOf test if the given String is a digest.
     * This tests the correct length and if all chars are in a hexadecimal range.
     * <p/>
     * The expectedSize must be the length of the digest in bytes, not in chars!
     *
     * @param digest       the digest !=null.
     * @param expectedSize the expectedSize of the digest in bytes as given by the algoritm
     *                     {@link java.security.MessageDigest#getDigestLength()}
     *                     Therewith this is the size in bytes, not the size of the String,
     *                     e.g. MD5 has a size of 16, and the String is tested for a length of 32.
     * @return boolean.
     */
    public static boolean isDigestOf(final String digest, final int expectedSize) {
        return digest.length() == (expectedSize << 1)
               && BYTE_HEX_PATTERN.matcher(digest).matches();
    }

    /**
     * Method isDigestOf test if the given byte[] array is a digest.
     * This simple tests for the correct length and is only supplied for convenience.
     *
     * @param digest    the digest !=null.
     * @param algorithm the algorithm !=null.
     * @return boolean.
     */
    public static boolean isDigestOf(final byte[] digest, final MessageDigest algorithm) {
        return isDigestOf(digest, algorithm.getDigestLength());
    }

    /**
     * Method isDigestOf test if the given char[] array is a digest.
     * This tests the correct length and if all chars are in a hexadecimal range.
     *
     * @param digest    the digest !=null.
     * @param algorithm the algorithm !=null.
     * @return boolean.
     */
    public static boolean isDigestOf(final char[] digest, final MessageDigest algorithm) {
        return isDigestOf(digest, algorithm.getDigestLength());
    }

    /**
     * Method isDigestOf test if the given char[] array is a digest.
     * This tests the correct length and if all chars are in a hexadecimal range.
     *
     * @param digest    the digest !=null.
     * @param algorithm the algorithm !=null.
     * @return boolean.
     */
    public static boolean isDigestOf(final String digest, final MessageDigest algorithm) {
        return isDigestOf(digest, algorithm.getDigestLength());
    }

    /**
     * Method isDigestOf test if the given byte[] array is a digest.
     * This simple tests for the correct length and is only supplied for convenience.
     *
     * @param digest    the digest !=null.
     * @param algorithm the algorithm !=null.
     * @return boolean.
     * @throws java.security.NoSuchAlgorithmException
     *          if algorithm not supported.
     */
    public static boolean isDigestOf(final byte[] digest, final String algorithm)
            throws NoSuchAlgorithmException {
        return isDigestOf(digest, MessageDigest.getInstance(algorithm).getDigestLength());
    }

    /**
     * Method isDigestOf test if the given char[] array is a digest.
     * This tests the correct length and if all chars are in a hexadecimal range.
     *
     * @param digest    the digest !=null.
     * @param algorithm the method !=null.
     * @return boolean.
     * @throws java.security.NoSuchAlgorithmException
     *          if algorithm not supported.
     */
    public static boolean isDigestOf(final char[] digest, final String algorithm)
            throws NoSuchAlgorithmException {
        return isDigestOf(digest, MessageDigest.getInstance(algorithm).getDigestLength());
    }

    /**
     * Method isDigestOf test if the given char[] array is a digest.
     * This tests the correct length and if all chars are in a hexadecimal range.
     *
     * @param digest    the digest !=null.
     * @param algorithm the method !=null.
     * @return boolean.
     * @throws java.security.NoSuchAlgorithmException
     *          if algorithm not supported.
     */
    public static boolean isDigestOf(final String digest, final String algorithm)
            throws NoSuchAlgorithmException {
        return isDigestOf(digest, MessageDigest.getInstance(algorithm).getDigestLength());
    }
}
