/****************************************************************
 SDI Security Services

 Copyright (C) 2004 by con terra GmbH

 Author: Marko Reiprecht

 Contact: Albert Remke, con terra GmbH, Martin-Luther-King-Weg 24,
 48155 Muenster, Germany, 52n@conterra.de

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 version 2 as published by the Free Software Foundation.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program (see gnu-gpl v2.txt); if not, write to
 the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 Boston, MA  02111-1307, USA or visit the web page of the Free
 Software Foundation, http://www.fsf.org.
 ****************************************************************/
/* $Id:StringInputStream.java 432 2007-06-19 13:18:47Z jessebluemr $ */
package org.n52.security.common.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CodingErrorAction;
import java.nio.charset.UnsupportedCharsetException;
import java.util.Arrays;

/**
 * Tasks of Class StringInputStream:
 * <ol>
 * <li>This is a Low Memory Implementation to convert a String in a InputStream</li>
 * <li>It converts only CHAR_BUFFER_SIZE chars at once</li>
 * </ol>
 *
 * @author reiprecht
 */
public class StringInputStream extends InputStream {

    private static final int CHAR_BUFFER_SIZE = 16;
    private String mContent;
    private int mIndex = -1;
    private CharsetEncoder mCharsetEncoder;
    private CharBuffer mCharBuffer;
    private ByteBuffer mByteBuffer;
    private boolean mClosed = false;

    public StringInputStream(String pContent, String pCharset) throws UnsupportedEncodingException {
        mContent = pContent;
        //Not available in JDK 1.4
        //if (pCharset == null) pCharset = Charset.defaultCharset().name();
        if (pCharset == null) pCharset = System.getProperty("file.encoding");
        try {
            mCharsetEncoder = Charset.forName(pCharset).newEncoder();
        } catch (UnsupportedCharsetException e) {
            throw new UnsupportedEncodingException(e.getCharsetName());
        }
        mCharsetEncoder.onMalformedInput(CodingErrorAction.REPLACE)
                .onUnmappableCharacter(CodingErrorAction.REPLACE);
        // Buffer Allocation
        mCharBuffer = CharBuffer.allocate(CHAR_BUFFER_SIZE);
        mByteBuffer = ByteBuffer.allocate(((int) mCharsetEncoder.maxBytesPerChar()) * CHAR_BUFFER_SIZE);
    }

    public int read() throws IOException {
        try {
            try {
                byte b = mByteBuffer.get();
                if (b != 0) return b;
            } catch (Exception e) {
            }
            mByteBuffer.position(0);
            Arrays.fill(mByteBuffer.array(), (byte) 0);
            mCharBuffer.position(0);
            fillCharBuffer();
            mCharBuffer.position(0);
            mCharsetEncoder.encode(mCharBuffer, mByteBuffer, false);
            mByteBuffer.position(0);

            byte b = mByteBuffer.get();
            if (b != 0) return b;
            return -1;
        } catch (IndexOutOfBoundsException e) {
            return -1;
        }
    }

    private void fillCharBuffer() {
        if (mClosed) throw new IndexOutOfBoundsException();
        int j = -1;
        try {
            while ((++j) < CHAR_BUFFER_SIZE) {
                mCharBuffer.put(mContent.charAt(++mIndex));
            }
        } catch (IndexOutOfBoundsException e) {
            mCharBuffer.limit(j);
            mClosed = true;
        }
    }

    public int available() throws IOException {
        return mContent.length();
    }

    public synchronized void reset() throws IOException {
        mClosed = false;
        mIndex = -1;
        mByteBuffer.clear();
        mCharBuffer.clear();
        Arrays.fill(mByteBuffer.array(), (byte) 0);
    }
}
