/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 22.03.2008
 ******************************************************************************/
/* $Id: $ */

package org.n52.security.common.xml;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.ref.WeakReference;
import java.net.URL;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;

/**
 * Class ClasspathEntityResolver tries to load external entities from the classpath.
 * <p/>
 * To find the entities in the classpath a mapping file must be specified.
 * The mapping file is located at "META-INF/xml/entity-mapping.properties".
 * <p/>
 * The file contains entries structured like this:
 * <p/>
 * <code>[publicId|systemId]=path in classpath</code>
 * <p/>
 * e.g.:
 * <p/>
 * <code>http://tempuri.org/test.dtd=META-INF/xml/dtds/test.dtd</code>
 *
 * @author Marko Reiprecht
 *         Created on 22.03.2008
 * @version $Revision: $
 */
public class ClasspathEntityResolver implements EntityResolver {

    // ------------------------------ FIELDS ------------------------------

    /**
     * The Logger.
     */
    private static final Log LOG = LogFactory.getLog(ClasspathEntityResolver.class);

    private static final byte[] EMPTY_BYTES = new byte[0];
    private static final String MAPPING_FILE_LOCATION = "META-INF/xml/entity-mapping.properties";
    private Map m_systemToClasspathMapping = new HashMap();

    private WeakReference m_classLoader;
    private boolean m_quite;

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Constructor creates a new instance in quite mode.
     */
    public ClasspathEntityResolver() {
        this(true);
    }

    /**
     * Constructor creates a new instance.
     *
     * @param quite the quite mode.
     */
    public ClasspathEntityResolver(final boolean quite) {
        this(quite, null);
    }

    /**
     * Constructor creates a new instance.
     *
     * @param quite  the quite mode.
     * @param loader the classloader to use, if null this.getClass().getClassLoader() is used.
     */
    public ClasspathEntityResolver(final boolean quite, final ClassLoader loader) {
        m_quite = quite;
        loadMapping(loader);
    }

    /**
     * Method loads all entries from the mapping files located at META-INF/xml/entity-mapping.properties.
     * A typical entity mapping looks like this:
     * <p/>
     * <code>[publicId|systemId]=path in classpath</code>
     * <p/>
     * e.g.:
     * <p/>
     * <code>http://tempuri.org/test.dtd=META-INF/xml/dtds/test.dtd</code>
     *
     * @param classloader the classloader if null this.getClass().getClassLoader() is used.
     */
    protected void loadMapping(final ClassLoader classloader) {
        ClassLoader resourceLoader = classloader;
        if (classloader == null) {
            resourceLoader = this.getClass().getClassLoader();
        }
        try {
            Enumeration files = resourceLoader.getResources(MAPPING_FILE_LOCATION);
            while (files.hasMoreElements()) {
                URL resource = (URL) files.nextElement();
                Properties mappings = new Properties();
                InputStream in = resource.openStream();
                try {
                    mappings.load(resource.openStream());
                    m_systemToClasspathMapping.putAll(mappings);
                } catch (IOException ex) {
                    if (LOG.isWarnEnabled()) {
                        LOG.warn(
                                "could not load <" + resource + ">. "
                                + "Loading of entities specified within this mapping file will fail. Error reason:",
                                ex);
                    }
                } finally {
                    try {
                        in.close();
                    } catch (IOException ex) {}
                }
            }
        } catch (IOException e) {
            if (LOG.isWarnEnabled()) {
                LOG.warn(
                        "could not load <entity-mapping.properties> from classpath, "
                        + "loading of entities will fail. Error reason:",
                        e);
            }
        }
        // store classloader as weak reference, otherwise this class is never garbaged.
        m_classLoader = new WeakReference(resourceLoader);
    }

    // --------------------- GETTER / SETTER METHODS ---------------------

    /**
     * Method gets the quite state.
     * If the EnityReslover is in quite mode, all non resolved entities are simply ignored (not loaded),
     * otherwise the parsers default mode is used.
     *
     * @return boolean.
     */
    public boolean isQuite() {
        return m_quite;
    }

    // ------------------------ INTERFACE METHODS ------------------------


    // --------------------- Interface EntityResolver ---------------------

    /**
     * Method resolves an entity by looking in a special mapping file
     * located at "META-INF/xml/entity-mapping.properties".
     * <p/>
     * First it looks for a publicId mapping and then for a systemId mapping.
     * If none mapping is found it returns a "emtpy" inputsource in quite mode or null in nonquite mode.
     *
     * @param publicId the publicId.
     * @param systemId the systemId.
     * @return InputSource.
     */
    public InputSource resolveEntity(final String publicId, final String systemId) {
        if (LOG.isTraceEnabled()) {
            LOG.trace("ClasspathEntityResolver.resolveEntity(" + publicId + "," + systemId + ")");
        }
        String resourcefile = null;
        if (publicId != null) {
            resourcefile = getMappedResourceFileName(publicId);
        }
        if (resourcefile == null && systemId != null) {
            resourcefile = getMappedResourceFileName(systemId);
        }
        if (resourcefile != null) {
            InputSource resolvedsource = createInputSource(resourcefile);
            if (resolvedsource != null) {
                return resolvedsource;
            }
        }
        // in quite mode simple disable the external loading,
        // otherwise use default parser behavior (e.g. make uri connection)
        return isQuite() ? emptyInputSource() : null;
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Method creates an input source for the given resource file.
     * <p/>
     * It tries to load the resourcefile from the classpath. If it fails it simple ignores the entity.
     *
     * @param resourcefile the resourcefile.
     * @return InputSource.
     */
    protected InputSource createInputSource(final String resourcefile) {
        if (LOG.isDebugEnabled()) { LOG.debug("Try to resolve entity from resource file <" + resourcefile + ">"); }
        ClassLoader loader = (ClassLoader) m_classLoader.get();
        // if classloader was garbaged, the resource can't be loaded
        if (loader != null && resourcefile != null) {
            InputStream resourcestream = loader.getResourceAsStream(resourcefile);
            if (resourcestream != null) {
                return new InputSource(resourcestream);
            } else {
                if (LOG.isWarnEnabled()) {
                    LOG.warn(
                            "entity mapping with value <" + resourcefile + "> seems broken,"
                            + " because the resourcestream does not exist in the classpath."
                            + " Entity won't be resolved.");
                }
            }
        }
        return null;
    }

    /**
     * Method creates an InputSource which points to an empty resource.
     *
     * @return InputSource.
     */
    protected InputSource emptyInputSource() {
        return new InputSource(new ByteArrayInputStream(EMPTY_BYTES));
    }

    /**
     * Method gets the mapped resource file name for the given identifier.
     *
     * @param identifier the identifier (e.g. publicId or systemId).
     * @return String if mapping is found, otherwise null.
     */
    protected String getMappedResourceFileName(final String identifier) {
        return (String) m_systemToClasspathMapping.get(identifier);
    }
}
