/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 24.03.2008
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.common.xml;

import javax.xml.parsers.DocumentBuilder;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.pool.KeyedPoolableObjectFactory;
import org.apache.commons.pool.impl.GenericKeyedObjectPool;

/**
 * Class DocumentBuilderPool implements a pool for document builders.
 * <p/>
 * The class makes use of a GenericKeyedObjectPool, which comes out of the commons-pool library,
 * to implement the parser pool.
 * <p/>
 * Currently the pooling settings are hard coded, e.g. the total size and max idle count,
 * but the settings shall meet the most requirements.
 *
 * @author Marko Reiprecht
 *         Created on 24.03.2008
 * @version $Revision: $
 */
public class DefaultDocumentBuilderPool implements DocumentBuilderPool {

    // ------------------------------ FIELDS ------------------------------

    /**
     * The Logger.
     */
    private static final Log LOG = LogFactory.getLog(DefaultDocumentBuilderPool.class);

    private GenericKeyedObjectPool m_documentBuilderPool;

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Constructor creates a new instance.
     */
    public DefaultDocumentBuilderPool() {
        GenericKeyedObjectPool.Config config = new GenericKeyedObjectPool.Config();
        // max number of total created documentbuilders for all possible options.
        config.maxTotal = 20;
        // max number of idle builders per option
        config.maxIdle = 5;
        // max number of active builder per option
        config.maxActive = 20;
        // don't grow, fail, if the pool is exhausted.
        // Blocking can lead to bad error states, so this poll will fail,
        // if more than 20 parsers are parallel in use.
        config.whenExhaustedAction = GenericKeyedObjectPool.WHEN_EXHAUSTED_GROW;
        // but only wait for maximal one second befor throw an exception
        //config.maxWait = 1000;

        // pool parsers
        m_documentBuilderPool = new GenericKeyedObjectPool(
                new KeyedPoolableDocumentBuilderFactory(), config);
    }

    // ------------------------ INTERFACE METHODS ------------------------


    // --------------------- Interface DocumentBuilderPool ---------------------

    /**
     * @see DocumentBuilderPool#acquire(DOMParserOptions)
     */
    public DocumentBuilder acquire(final DOMParserOptions options) {
        try {
            return (DocumentBuilder) m_documentBuilderPool.borrowObject(options);
        } catch (Exception e) {
            throw new DOMParserException("DocumentBuilder could not retrieved from pool", e);
        }
    }

    /**
     * @see DocumentBuilderPool#release(DocumentBuilder, DOMParserOptions)
     */
    public void release(final DocumentBuilder builder, final DOMParserOptions options) {
        try {
            m_documentBuilderPool.returnObject(options, builder);
        } catch (Exception e) {
            if (LOG.isWarnEnabled()) { LOG.warn("Error on return pooled document builder", e); }
        }
    }

    // -------------------------- INNER CLASSES --------------------------

    /**
     * Class KeyedPoolableDocumentBuilderFactory creates instances of DocumentBuilder if required.
     *
     * @author Marko Reiprecht
     *         Created on 22.04.2008
     * @version $Revision: $
     */
    private static class KeyedPoolableDocumentBuilderFactory implements KeyedPoolableObjectFactory {

        /**
         * Method makeObject, creates a new DocumentBuilder, the key is a DOMParserOptions instance.
         *
         * @param key the key a DOMParserOptions instance.
         * @return Object a new DocumentBuilder.
         * @throws Exception if error occurs.
         */
        public Object makeObject(final Object key) throws Exception {
            if (LOG.isTraceEnabled()) {
                LOG.trace("DefaultDocumentBuilderPool$PooledDocumentBuilderFactory.makeObject(" + key + ")");
            }
            return DOMParser.createDocumentBuilder((DOMParserOptions) key);
        }

        /**
         * Method destroyObject, does nothing.
         *
         * @param key the key a DOMParserOptions instance.
         * @param o   the o a DocumentBuilder.
         * @throws Exception if error occurs.
         */
        public void destroyObject(final Object key, final Object o) throws Exception {
            if (LOG.isTraceEnabled()) {
                LOG.trace("DefaultDocumentBuilderPool$PooledDocumentBuilderFactory.makeObject(" + key + ")");
            }
            // nothing to do here
        }

        /**
         * Method validateObject, does nothing, DocumentBuilders are always valid.
         *
         * @param key the key a DOMParserOptions instance.
         * @param o   the o a DocumentBuilder.
         * @return boolean always true.
         */
        public boolean validateObject(final Object key, final Object o) {
            // nothing to do here
            return true;
        }

        /**
         * Method activateObject, "activates" the documentBuilder, this resets the buidler to default state.
         *
         * @param key the key a DOMParserOptions instance.
         * @param o   the o a DocumentBuilder.
         * @throws Exception if error occurs.
         */
        public void activateObject(final Object key, final Object o) throws Exception {
            ((DocumentBuilder) o).setErrorHandler(null);
            ((DocumentBuilder) o).setEntityResolver(DOMParser.getDefaultEntityResolver());
        }

        /**
         * Method passivateObject, resets the ErrorHandler and EntityReslover states.
         *
         * @param key the key a DOMParserOptions instance.
         * @param o   the o a DocumentBuilder.
         * @throws Exception if error occurs.
         */
        public void passivateObject(final Object key, final Object o) throws Exception {
            ((DocumentBuilder) o).setErrorHandler(null);
            ((DocumentBuilder) o).setEntityResolver(null);
        }
    }
}