/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 18.04.2008
 ******************************************************************************/
package org.n52.security.common.xml;

import java.io.IOException;
import java.io.Writer;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Characters;
import javax.xml.stream.events.EndElement;
import javax.xml.stream.events.Namespace;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;
import org.xml.sax.InputSource;

/**
 * Class XMLStreamChunkReader reads an xml document with an internal xml stream reader
 * and as chunks and calls a subsets of XMLStreamChunkHandler instances.
 * <p/>
 * Xml elements for which no handler is defined are simply skipped.
 *
 * @author Marko Reiprecht
 *         Created on 18.04.2008
 * @version $Revision: $
 */
public class XMLChunkReader {

    // ------------------------------ FIELDS ------------------------------

    private HashMap m_partialStreamHandler = new HashMap();

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Method registers a PartialStreamHandler for a Qname.
     *
     * @param qname   the qname.
     * @param handler the handler.
     * @return this.
     */
    public XMLChunkReader addHandler(final QName qname, final XMLChunkHandler handler) {
        m_partialStreamHandler.put(qname, handler);
        return this;
    }

    /**
     * Method gets the Handler for the qname.
     *
     * @param qname the qname.
     * @return XmlPartialStreamHandler.
     */
    public XMLChunkHandler getHandler(final QName qname) {
        return (XMLChunkHandler) m_partialStreamHandler.get(qname);
    }

    /**
     * Method clears all Handlers (removes them).
     */
    public void clearHandlers() {
        m_partialStreamHandler.clear();
    }

    /**
     * Method reads an xml source.
     *
     * @param source the source.
     * @throws XMLStreamException if error occurs.
     */
    public void read(final InputSource source) throws XMLStreamException {
        // Create the input reader to read the source.
        XMLInputFactory inFactory = XMLInputFactory.newInstance();
        inFactory.setProperty(XMLInputFactory.IS_NAMESPACE_AWARE, Boolean.TRUE);
        // in jdk 1.6 this method is throws an Malformed URL exception, so we must use
        // the plain methods
        //XMLEventReader reader = inFactory.createXMLEventReader(new SAXSource(source));
        XMLEventReader reader;
        if (source.getByteStream() != null) {
            if (source.getEncoding() != null && source.getEncoding().length() > 0) {
                reader = inFactory.createXMLEventReader(source.getByteStream(), source.getEncoding());
            } else if (source.getSystemId() != null && source.getSystemId().length() > 0) {
                reader = inFactory.createXMLEventReader(source.getSystemId(), source.getByteStream());
            } else {
                reader = inFactory.createXMLEventReader(source.getByteStream());
            }
        } else if (source.getCharacterStream() != null) {
            if (source.getSystemId() != null && source.getSystemId().length() > 0) {
                reader = inFactory.createXMLEventReader(source.getSystemId(), source.getCharacterStream());
            } else {
                reader = inFactory.createXMLEventReader(source.getCharacterStream());
            }
        } else {
            throw new IllegalArgumentException("InputSource holds no valid input");
        }

        XMLEventFactory fac = XMLEventFactory.newInstance();

        // create a dynamicaly grown writer pool
        LinkedList writerPool = new LinkedList();
        try {
            LinkedList handlerStack = new LinkedList();
            LinkedList namespaceStack = new LinkedList();

            while (reader.hasNext()) {
                XMLEvent evt = reader.nextEvent();
                // If this event starts the element of interest,
                // change the flag to indicate such.
                if (evt.isStartElement()) {
                    StartElement elem = (StartElement) evt;
                    NamespaceStackEntry namespaceEntry = new NamespaceStackEntry(elem.getNamespaces(), namespaceStack);
                    QName qName = elem.getName();
                    XMLChunkHandler handler = getHandler(qName);
                    if (handler != null) {
                        if (!handlerStack.isEmpty()) {
                            HandlerStackEntry entry = (HandlerStackEntry) handlerStack.getFirst();
                            entry.getHandler().handleChunkBeforChildHandler(entry.getQname(),
                                                                            entry.getWriterContent(),
                                                                            qName, handler);
                        }
                        HandlerStackEntry newActive = new HandlerStackEntry(handler, qName, writerPool);
                        handlerStack.addFirst(newActive);
                        newActive.getHandler().handleChunkStart(qName);
                        // we have to append the start element with all available namespaces
                        // otherwise the element is not correct
                        // and XMLOutputFactory.IS_REPAIRING_NAMESPACES, does not work,
                        // because some elements are skiped
                        StartElement elemCorrected = fac.createStartElement(elem.getName(),
                                                                            elem.getAttributes(),
                                                                            namespaceEntry.getNamespaces());
                        newActive.getEventWriter().add(elemCorrected);
                        continue;
                    }
                }
                // If we are in the element, write the events
                // back out.
                if (!handlerStack.isEmpty()) {
                    if (evt.isCharacters() && ((Characters) evt).isWhiteSpace()) {
                        continue;
                    }
                    ((HandlerStackEntry) handlerStack.getFirst()).getEventWriter().add(evt);
                }
                // If this is the end of the element
                // and an handler is registered for this element, so call it
                if (evt.isEndElement()) {
                    namespaceStack.removeFirst();
                    if (!handlerStack.isEmpty()) {
                        EndElement elem = (EndElement) evt;
                        HandlerStackEntry handlerentry = (HandlerStackEntry) handlerStack.getFirst();
                        if (elem.getName().equals(handlerentry.getQname())) {
                            handlerStack.removeFirst();
                            handlerentry.getHandler().handleChunkEnd(handlerentry.getQname(),
                                                                     handlerentry.getWriterContent());

                            if (!handlerStack.isEmpty()) {
                                HandlerStackEntry parentEntry = (HandlerStackEntry) handlerStack.getFirst();
                                parentEntry.getHandler().handleChunkAfterChildHandler(parentEntry.getQname(),
                                                                                      parentEntry.getWriterContent(),
                                                                                      handlerentry.getQname(),
                                                                                      handlerentry.getHandler());
                            }
                            // release the used writer
                            handlerentry.release();
                        }
                    }
                }
            }
        } finally {
            reader.close();
        }
    }

    // -------------------------- INNER CLASSES --------------------------

    /**
     * Class NamespaceStackEntry encapsulates the namespaces available for each element.
     * The namespaces must be managed in stack and appended to the startelement of each chunk,
     * so that the xml is legal.
     *
     * @author Marko Reiprecht
     *         Created on 19.04.2008
     * @version $Revision: $
     */
    private static class NamespaceStackEntry {

        private Set m_namespaces;

        /**
         * Constructor creates a new instance.
         *
         * @param elementNamespaces the elementNamespaces.
         * @param namespaceStack    the namespaceStack.
         */
        private NamespaceStackEntry(final Iterator elementNamespaces,
                                    final LinkedList namespaceStack) {
            m_namespaces = new HashSet();
            if (!namespaceStack.isEmpty()) {
                if (elementNamespaces.hasNext()) {
                    // there will some namespaces appended, so we have to copy the current namespaces
                    m_namespaces.addAll(((NamespaceStackEntry) namespaceStack.getFirst()).m_namespaces);
                } else {
                    // don't copy only link, because it are the same namespaces
                    m_namespaces = ((NamespaceStackEntry) namespaceStack.getFirst()).m_namespaces;
                }
            }
            while (elementNamespaces.hasNext()) {
                Namespace namespace = (Namespace) elementNamespaces.next();
                m_namespaces.add(namespace);
            }
            namespaceStack.addFirst(this);
        }

        /**
         * Method gets the namespaces.
         *
         * @return Iterator.
         */
        Iterator getNamespaces() {
            return m_namespaces.iterator();
        }
    }

    /**
     * Class WriterPoolEntry encapsulates eventwriters, which are used to generate the xmlchunks for each handler.
     * They can be reused for different handlers, therefor its structured as a pool entry.
     *
     * @author Marko Reiprecht
     *         Created on 19.04.2008
     * @version $Revision: $
     */
    private static class WriterPoolEntry {

        private ResettableStringWriter m_stringWriter;
        private XMLEventWriter m_eventWriter;
        private LinkedList m_poolQueue;

        /**
         * Constructor creates a new instance.
         *
         * @param poolQueue the poolQueue.
         * @throws XMLStreamException if error occurs.
         */
        private WriterPoolEntry(final LinkedList poolQueue) throws XMLStreamException {
            // Create an output writer. The writer uses the resettable
            // buffer so we can use the same writer continuously.
            XMLOutputFactory outFactory = XMLOutputFactory.newInstance();
            // the flag is needed so that namespaces are appended.
            // the flag works only for element namespaces, not for attribute namespaces,
            // therefor the namespaces are managed in an extra stack.
            //outFactory.setProperty(XMLOutputFactory.IS_REPAIRING_NAMESPACES, Boolean.TRUE);
            m_stringWriter = new ResettableStringWriter();
            m_eventWriter = outFactory.createXMLEventWriter(m_stringWriter);
            m_poolQueue = poolQueue;
        }

        /**
         * Method gets the stringWriter.
         *
         * @return ResettableStringWriter.
         */
        public ResettableStringWriter getStringWriter() {
            return m_stringWriter;
        }

        /**
         * Method gets the eventWriter.
         *
         * @return XMLEventWriter.
         */
        public XMLEventWriter getEventWriter() {
            return m_eventWriter;
        }

        /**
         * Method release the pool entry and make it available for other handlers.
         */
        public void release() {
            m_stringWriter.reset();
            m_poolQueue.addLast(this);
        }
    }

    /**
     * Class HandlerStackEntry.
     *
     * @author Marko Reiprecht
     *         Created on 18.04.2008
     * @version $Revision: $
     */
    private static class HandlerStackEntry {

        private XMLChunkHandler m_handler;
        private QName m_qname;
        private WriterPoolEntry m_poolEntry;

        /**
         * Constructor creates a new instance.
         *
         * @param handler    the handler.
         * @param qname      the qname.
         * @param writerPool the pooled event writers.
         * @throws javax.xml.stream.XMLStreamException
         *          if a new writer pool instance can't be created.
         */
        private HandlerStackEntry(final XMLChunkHandler handler, final QName qname, final LinkedList writerPool)
                throws XMLStreamException {
            m_handler = handler;
            m_qname = qname;
            if (writerPool.isEmpty()) {
                m_poolEntry = new WriterPoolEntry(writerPool);
            } else {
                m_poolEntry = (WriterPoolEntry) writerPool.removeFirst();
            }
        }

        /**
         * Method gets the handler.
         *
         * @return XmlPartialStreamHandler.
         */
        public XMLChunkHandler getHandler() {
            return m_handler;
        }

        /**
         * Method gets the qname.
         *
         * @return QName.
         */
        public QName getQname() {
            return m_qname;
        }

        /**
         * Method gets the stringWriter.
         *
         * @return ResettableStringWriter.
         */
        public ResettableStringWriter getStringWriter() {
            return m_poolEntry.getStringWriter();
        }

        /**
         * Method gets the eventWriter.
         *
         * @return XMLEventWriter.
         */
        public XMLEventWriter getEventWriter() {
            return m_poolEntry.getEventWriter();
        }

        /**
         * Method gets the poolEntry.
         *
         * @return WriterPoolEntry.
         */
        public WriterPoolEntry getPoolEntry() {
            return m_poolEntry;
        }

        /**
         * Method release.
         */
        public void release() {
            m_poolEntry.release();
        }

        /**
         * Method gets the currently written content in the buffer.
         *
         * @return String.
         * @throws XMLStreamException if error occurs.
         */
        public String getWriterContent() throws XMLStreamException {
            m_poolEntry.getEventWriter().flush();
            return m_poolEntry.getStringWriter().getBuffer().substring(0);
        }
    }

    /**
     * A Writer that buffers the data in memory as a string. The buffer can be
     * reset at any time and the contents of the buffer will be returned.
     */
    private static class ResettableStringWriter extends Writer {

        /**
         * The internal buffer used to store the data written to this stream.
         */
        private StringBuffer m_buffer = new StringBuffer(1024);

        /**
         * The maximum number of chars that this writer will reuse.
         * (2MB)
         */
        private int m_maxChars = 1024 * 1024;

        /**
         * Constructor creates a new instance.
         */
        private ResettableStringWriter() {}

        /**
         * Constructor creates a new instance.
         *
         * @param maxChars the maxChars this buffer will reuse.
         */
        private ResettableStringWriter(final int maxChars) {
            m_maxChars = maxChars;
        }

        /**
         * Method close.
         *
         * @throws IOException if error occurs.
         */
        public void close() throws IOException {
            // no op
        }

        /**
         * Method flush.
         *
         * @throws IOException if error occurs.
         */
        public void flush() throws IOException {
            // no op
        }

        /**
         * Method write.
         *
         * @param cbuf the cbuf.
         * @param off  the off.
         * @param len  the len.
         * @throws IOException if error occurs.
         */
        public void write(final char[] cbuf, final int off, final int len) throws IOException {
            m_buffer.append(cbuf, off, len);
        }

        /**
         * Method gets the buffer.
         *
         * @return StringBuffer.
         */
        public StringBuffer getBuffer() {
            return m_buffer;
        }

        /**
         * Method sets the buffer.
         *
         * @param buffer the buffer.
         */
        public void setBuffer(final StringBuffer buffer) {
            m_buffer = buffer;
        }

        /**
         * Resets the stream to a length of zero. Any data written to the stream
         * since the last call to reset will be returned.
         *
         * @return the contents of the stream buffer
         */
        public ResettableStringWriter reset() {
            if (m_buffer.length() > m_maxChars) {
                m_buffer = new StringBuffer(1024);
            }
            m_buffer.setLength(0);
            return this;
        }
    }
}
