/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 10.07.2008
 * Modified: $Date $
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.common.xml;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;
import javax.xml.namespace.QName;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/**
 * Class XMLPath represents an XPath and it's evaluation.
 * <p/>
 * You create a "source" XMLPath by using the {@link XMLPathCtx#findIn} methods.
 * From there you can use the
 * <ul>
 * <li>{@link #node(String)}</li>
 * <li>{@link #all(String)}</li>
 * <li>{@link #bool(String)}</li>
 * <li>{@link #number(String)}</li>
 * <li>{@link #text(String)}</li>
 * </ul>
 * methods to build a whole xpath.
 * After creation of an whole xpath you can call the get() method
 * of each concrete XMLPath instance, to get the result of the query.
 * <p/>
 * Sample (finds a root node of a document):
 * <pre><code>
 * XMLPathCtx ctx = XMLPathCtx.createNew()
 *                     .addNamespace("t", "http://tempuri.org/test");
 * Node result = ctx.findIn(document).node("/*").get();
 * </code></pre>
 *
 * @author Marko Reiprecht
 *         Created on 10.07.2008
 * @version $Revision: $
 */
public abstract class XMLPath {

    /**
     * The Logger.
     */
    private static final Log LOG = LogFactory.getLog(XMLPath.class);

    protected XMLPathCtx m_ctx;
    protected XMLPath m_previous;
    protected XPathExpression m_xPath;


    /**
     * Constructor creates a new instance.
     *
     * @param ctx the ctx.
     */
    protected XMLPath(final XMLPathCtx ctx) {
        m_ctx = ctx;
    }

    /**
     * Constructor creates a new instance.
     *
     * @param previous the previous.
     */
    protected XMLPath(final XMLPath previous) {
        m_ctx = previous.m_ctx;
        m_previous = previous;
    }

    /**
     * Method evals the xpath.
     *
     * @return Object.
     */
    protected Object eval() {
        try {
            // xpath.evaluate supports only a instance of "Node" as first parameter
            // so to ensure correct use of the classes here a check of the input is done
            if (XPathConstants.STRING == m_previous.getReturnType()) {
                // evaluate a previous string
                // wrap the previous string with an InputSource (so the string must be xml)
                if (LOG.isWarnEnabled()) {
                    LOG.warn("you use an xpath expression on an previous string result, " +
                             "this works only if the string is wellformed xml");
                }
                return m_xPath.evaluate(new InputSource(new StringReader((String) m_previous.eval())), getReturnType());
            } else if (XPathConstants.NODE == m_previous.getReturnType()) {
                // evaluate a previous node
                return m_xPath.evaluate(m_previous.eval(), getReturnType());
            } else if (XPathConstants.NODESET == m_previous.getReturnType()) {
                // evaluate a previous node set
                return evalNodeList((NodeList) m_previous.eval());
            } else if (m_previous.getReturnType() == null) {
                // evaluate a previous input source
                return m_xPath.evaluate((InputSource) m_previous.eval(), getReturnType());
            } else if (XPathConstants.NUMBER == m_previous.getReturnType()) {
                // evaluate a previous number
                throw new IllegalStateException("you can not uses an xpath expression on an previous number result");
            } else if (XPathConstants.BOOLEAN == m_previous.getReturnType()) {
                // evaluate a previous boolean
                throw new IllegalStateException("you can not uses an xpath expression on an previous boolean result");
            } else {
                throw new IllegalStateException(
                        "return type of unknown format expected <" + m_previous.getReturnType() + ">");
            }
        } catch (XPathExpressionException e) {
            throw new XMLPathException(e);
        }
    }

    /**
     * Method evals a path on a NodeList.
     *
     * @param nodeList the nodeList.
     * @return Object.
     * @throws XPathExpressionException if error occurs.
     */
    private Object evalNodeList(final NodeList nodeList) throws XPathExpressionException {
        if (getReturnType() != XPathConstants.NODESET) {
            for (int i = 0, l = nodeList.getLength(); i < l; ++i) {
                Object result = m_xPath.evaluate(nodeList.item(i), getReturnType());
                if (result != null) {
                    return result;
                }
            }
            return null;
        } else {
            SimpleNodeList resultNodes = new SimpleNodeList();
            for (int i = 0, l = nodeList.getLength(); i < l; ++i) {
                resultNodes.addAll((NodeList) m_xPath.evaluate(nodeList.item(i), getReturnType()));
            }
            return resultNodes;
        }
    }

    /**
     * Method gets the returnType of an evaluation.
     *
     * @return QName.
     */
    protected abstract QName getReturnType();

    /**
     * Method compiles an xpath.
     *
     * @param xpath the xpath.
     * @return XMLPath.
     */
    protected XMLPath compile(final String xpath) {
        m_xPath = m_ctx.lookUp(xpath);
        return this;
    }


    /**
     * Method node, expects a xpath query resulting in a node.
     *
     * @param xpath the xpath.
     * @return NodeXMLPath.
     */
    public NodeXMLPath node(final String xpath) {
        return (NodeXMLPath) new NodeXMLPath(this).compile(xpath);
    }

    /**
     * Method all, expects a xpath query resulting in a NodeList.
     *
     * @param xpath the xpath.
     * @return NodeSetXMLPath.
     */
    public NodeSetXMLPath all(final String xpath) {
        return (NodeSetXMLPath) new NodeSetXMLPath(this).compile(xpath);
    }

    /**
     * Method number, expects a xpath query resulting in a number.
     *
     * @param xpath the xpath.
     * @return NumberXMLPath.
     */
    public NumberXMLPath number(final String xpath) {
        return (NumberXMLPath) new NumberXMLPath(this).compile(xpath);
    }

    /**
     * Method text, expects a xpath query resulting in a string.
     *
     * @param xpath the xpath.
     * @return StringXMLPath.
     */
    public StringXMLPath text(final String xpath) {
        return (StringXMLPath) new StringXMLPath(this).compile(xpath);
    }

    /**
     * Method bool, expects a xpath query resulting in a Boolean.
     *
     * @param xpath the xpath.
     * @return BooleanXMLPath.
     */
    public BooleanXMLPath bool(final String xpath) {
        return (BooleanXMLPath) new BooleanXMLPath(this).compile(xpath);
    }

    /**
     * Class SourceXMLPath, represents the xpath root/or source of evaluation.
     *
     * @author Marko Reiprecht
     *         Created on 16.07.2008
     * @version $Revision: $
     */
    public static class SourceXMLPath extends XMLPath {

        private Object m_source;
        private QName m_returnType;

        /**
         * Constructor creates a new instance.
         *
         * @param ctx the ctx.
         * @param doc the doc.
         */
        protected SourceXMLPath(final XMLPathCtx ctx, final Document doc) {
            super(ctx);
            m_source = doc;
            m_returnType = XPathConstants.NODE;
        }

        /**
         * Constructor creates a new instance.
         *
         * @param ctx the ctx.
         * @param doc the doc.
         */
        protected SourceXMLPath(final XMLPathCtx ctx, final Node doc) {
            super(ctx);
            m_source = doc;
            m_returnType = XPathConstants.NODE;
        }

        /**
         * Constructor creates a new instance.
         *
         * @param ctx the ctx.
         * @param set the set.
         */
        protected SourceXMLPath(final XMLPathCtx ctx, final NodeList set) {
            super(ctx);
            m_source = set;
            m_returnType = XPathConstants.NODESET;
        }

        /**
         * Constructor creates a new instance.
         *
         * @param ctx    the ctx.
         * @param source the source.
         */
        protected SourceXMLPath(final XMLPathCtx ctx, final InputSource source) {
            super(ctx);
            m_source = source;
            m_returnType = null;
        }

        /**
         * Constructor creates a new instance.
         *
         * @param ctx    the ctx.
         * @param source the source.
         */
        protected SourceXMLPath(final XMLPathCtx ctx, final String source) {
            super(ctx);
            m_source = source;
            m_returnType = XPathConstants.STRING;
        }

        /**
         * @see XMLPath#eval()
         */
        protected Object eval() {
            return m_source;
        }

        /**
         * @see XMLPath#getReturnType()
         */
        protected QName getReturnType() {
            return m_returnType;
        }
    }

    /**
     * Class NodeXMLPath, represents xpaths resulting in nodes.
     *
     * @author Marko Reiprecht
     *         Created on 16.07.2008
     * @version $Revision: $
     */
    public static class NodeXMLPath extends XMLPath {

        /**
         * Constructor creates a new instance.
         *
         * @param previous the previous.
         */
        protected NodeXMLPath(final XMLPath previous) {
            super(previous);
        }

        /**
         * Method gets the queried node.
         *
         * @return Node.
         */
        public Node get() {
            return (Node) eval();
        }

        /**
         * @see XMLPath#getReturnType()
         */
        protected QName getReturnType() {
            return XPathConstants.NODE;
        }
    }

    /**
     * Class NodeSetXMLPath, represents xpaths resulting in NodeLists.
     *
     * @author Marko Reiprecht
     *         Created on 16.07.2008
     * @version $Revision: $
     */
    public static class NodeSetXMLPath extends XMLPath {

        /**
         * Constructor creates a new instance.
         *
         * @param previous the previous.
         */
        protected NodeSetXMLPath(final XMLPath previous) {
            super(previous);
        }

        /**
         * Method gets the queried NodeList.
         *
         * @return NodeList.
         */
        public NodeList get() {
            return (NodeList) eval();
        }

        /**
         * @see XMLPath#getReturnType()
         */
        protected QName getReturnType() {
            return XPathConstants.NODESET;
        }
    }

    /**
     * Class StringXMLPath, represents xpaths queries resulting in string values.
     *
     * @author Marko Reiprecht
     *         Created on 16.07.2008
     * @version $Revision: $
     */
    public static class StringXMLPath extends XMLPath {

        /**
         * Constructor creates a new instance.
         *
         * @param previous the previous.
         */
        protected StringXMLPath(final XMLPath previous) {
            super(previous);
        }

        /**
         * Method gets the String, queried by the path.
         *
         * @return String.
         */
        public String get() {
            return (String) eval();
        }

        /**
         * @see XMLPath#getReturnType()
         */
        protected QName getReturnType() {
            return XPathConstants.STRING;
        }
    }

    /**
     * Class NumberXMLPath, represents xpath queries resulting in number values.
     *
     * @author Marko Reiprecht
     *         Created on 16.07.2008
     * @version $Revision: $
     */
    public static class NumberXMLPath extends XMLPath {

        /**
         * Constructor creates a new instance.
         *
         * @param previous the previous.
         */
        protected NumberXMLPath(final XMLPath previous) {
            super(previous);
        }

        /**
         * Method gets the queried number.
         *
         * @return Number.
         */
        public Number get() {
            return (Number) eval();
        }

        /**
         * @see XMLPath#getReturnType()
         */
        protected QName getReturnType() {
            return XPathConstants.NUMBER;
        }
    }

    /**
     * Class BooleanXMLPath, represents xpath queries resulting in boolean values.
     *
     * @author Marko Reiprecht
     *         Created on 16.07.2008
     * @version $Revision: $
     */
    public static class BooleanXMLPath extends XMLPath {

        /**
         * Constructor creates a new instance.
         *
         * @param previous the previous.
         */
        protected BooleanXMLPath(final XMLPath previous) {
            super(previous);
        }

        /**
         * Method gets the queried boolean value.
         *
         * @return Boolean.
         */
        public Boolean get() {
            return (Boolean) eval();
        }

        /**
         * @see XMLPath#getReturnType()
         */
        protected QName getReturnType() {
            return XPathConstants.BOOLEAN;
        }
    }

    /**
     * Class SimpleNodeList implements the NodeList interface.
     * It is used to make xpath queries of NodeLists possible.
     *
     * @author Marko Reiprecht
     *         Created on 16.07.2008
     * @version $Revision: $
     */
    private static class SimpleNodeList implements NodeList {

        private List m_nodes = new ArrayList();


        /**
         * Method gets the length.
         *
         * @return int.
         */
        public int getLength() {
            return m_nodes.size();
        }

        /**
         * Method item.
         *
         * @param index the index.
         * @return Node.
         */
        public Node item(final int index) {
            return (Node) m_nodes.get(index);
        }

        /**
         * Method addAll.
         *
         * @param result the result.
         */
        public void addAll(final NodeList result) {
            for (int i = 0, l = result.getLength(); i < l; ++i) {
                add(result.item(i));
            }
        }

        /**
         * Method add.
         *
         * @param node the node.
         */
        public void add(final Node node) {
            m_nodes.add(node);
        }
    }
}
