/*******************************************************************************
 * Copyright  2007 52North Initiative for Geospatial Open Source Software GmbH
 *
 * Author: Marko Reiprecht, con terra GmbH
 *
 * Contact: Andreas Wytzisk, 52North Initiative for Geospatial Open Source
 * Software GmbH, Martin-Luther-King-Weg 24, 48155 Muenster, Germany,
 * info@52north.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; even without the implied WARRANTY OF MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundations web page, http://www.fsf.org.
 *
 ******************************************************************************/
/* $Id: Obligation.java 1149 2008-08-05 11:48:03Z MarkoReiprecht $ */
package org.n52.security.decision;

import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import org.n52.security.common.attributes.Attribute;
import org.n52.security.common.attributes.AttributeUtil;
import org.n52.security.common.attributes.Attributes;

/**
 * This class represents an obligation.
 * An obligation could be set in an PDPResponse by an PDPProxy.
 * Each obligation HAS to be fulfilled.
 *
 * @author Bastian Schaeffer
 * @author Marko Reiprecht
 * @version $Revision: 1149 $
 */
public class Obligation implements Serializable {

    // ------------------------------ FIELDS ------------------------------

    /**
     * Field FULLFILL_ON_PERMIT represents the 'permit' state of an obligation.
     */
    public static final int FULLFILL_ON_PERMIT = 1;
    /**
     * Field FULLFILL_ON_DENY  represents the 'deny' state of an obligation.
     */
    public static final int FULLFILL_ON_DENY = 2;

    /**
     * Field serialVersionUID.
     */
    private static final long serialVersionUID = 8242177586880892678L;

    /**
     * Field fullFillOn  olds the fullfill on state.
     * It must have the value {@link #FULLFILL_ON_PERMIT} or {@link #FULLFILL_ON_DENY}
     */
    protected int m_fullFillOn;
    /**
     * Field id carries the id of this obligation.
     * The id must only be unique whithin the scope of an PDPResponse.
     */
    protected String m_id;
    /**
     * Field attributes carries a collection of attributes with type {@link org.n52.security.common.attributes.Attribute}.
     */
    protected Attributes m_attributes;

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * Creates an obligation.
     * The attributes parameter is internal copied
     * and attributes with duplicated keys are not supported.
     * The objects in attributes must be of type {@link Attribute}.
     *
     * @param fullFillOn the decision on which the obligation has to be fulfilled.
     * @param id         the id of the obligation.
     * @param attributes a collection of attributes.
     * @throws IllegalArgumentException if fullFillOn is out of range,
     *                                  or the attributes are not subclasses of {@link Attribute}
     * @throws NullPointerException     if id or attributes is null.
     */
    public Obligation(final int fullFillOn, final String id, final Collection attributes)
            throws NullPointerException, IllegalArgumentException {
        checkFullFillOn(fullFillOn);
        if (id == null) {
            throw new IllegalArgumentException("<id> must not null");
        }
        if (attributes == null) {
            throw new IllegalArgumentException("<attributes> must not null");
        }

        this.m_fullFillOn = fullFillOn;
        this.m_id = id;
        this.m_attributes = new Attributes(attributes);
    }

    /**
     * Checks the fullfill on parameter in the constructor.
     * Subclasses can override this method to support other full fill on flags.
     *
     * @param fullFillOn the full fill on parameter
     * @throws IllegalArgumentException if fullFillOn is out of range.
     * @see #Obligation(int,String,java.util.Collection)
     */
    protected void checkFullFillOn(final int fullFillOn) throws IllegalArgumentException {
        switch (fullFillOn) {
            case FULLFILL_ON_PERMIT:
            case FULLFILL_ON_DENY:
                break;
            default:
                throw new IllegalArgumentException("<fullFillOn> has to be <Obligation.FULLFILL_ON_PERMIT> or <Obligation.FULLFILL_ON_DENY>");
        }
    }

    // --------------------- GETTER / SETTER METHODS ---------------------

    /**
     * @return Returns the fullFillOn Flag.
     * @see #FULLFILL_ON_DENY
     * @see #FULLFILL_ON_PERMIT
     */
    public int getFullFillOn() {
        return m_fullFillOn;
    }

    /**
     * @return Returns the id.
     */
    public String getId() {
        return m_id;
    }

    /**
     * Method gets the fullFillOnPermit.
     *
     * @return boolean.
     */
    public boolean isFullFillOnPermit() {
        return m_fullFillOn == FULLFILL_ON_PERMIT;
    }

    /**
     * Method gets the fullFillOnDeny.
     *
     * @return boolean.
     */
    public boolean isFullFillOnDeny() {
        return m_fullFillOn == FULLFILL_ON_DENY;
    }

    // ------------------------ CANONICAL METHODS ------------------------

    /**
     * Method equals checks for value equality of two obligation objects.
     *
     * @param o of type Object
     * @return boolean
     */
    public boolean equals(final Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        final Obligation that = (Obligation) o;

        if (m_fullFillOn != that.m_fullFillOn) {
            return false;
        }
        if (!m_attributes.equals(that.m_attributes)) {
            return false;
        }
        if (!m_id.equals(that.m_id)) {
            return false;
        }

        return true;
    }

    /**
     * Method hashCode generates a hashcode for this obligation.
     *
     * @return int
     */
    public int hashCode() {
        int result;
        result = m_fullFillOn;
        result = 31 * result + m_id.hashCode();
        result = 31 * result + m_attributes.hashCode();
        return result;
    }

    /**
     * Method toString generates a simple object info string, used only for logging purpose.
     *
     * @return String
     */
    public String toString() {
        final StringBuffer sb = new StringBuffer();
        sb.append("org.n52.security.decision.Obligation");
        sb.append("{fullFillOn=").append(getFullFillOn());
        sb.append(", id='").append(getId()).append('\'');
        sb.append(", attributes=").append(getAttributes());
        sb.append('}');
        return sb.toString();
    }

    /**
     * Gets the attributes.
     * The collection is unmodifiable.
     *
     * @return returns the attributes.
     */
    public Collection getAttributes() {
        return m_attributes.getAttributes();
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Returns true if an attribute with the given key is present.
     *
     * @param key a key, asserted to be not null.
     * @return true if attribute with key is present.
     */
    public boolean containsAttribute(final String key) {
        return m_attributes.containsAttribute(key);
    }

    /**
     * Gets an Attribute for a given key.
     *
     * @param key the key of the attribute.
     * @return the found attribute or null if nothing is found.
     */
    public Attribute getAttribute(final String key) {
        return m_attributes.getAttribute(key);
    }
}