/*******************************************************************************
 * Copyright  2007 52North Initiative for Geospatial Open Source Software GmbH
 *
 * Author: Marko Reiprecht, con terra GmbH
 *
 * Contact: Andreas Wytzisk, 52North Initiative for Geospatial Open Source
 * Software GmbH, Martin-Luther-King-Weg 24, 48155 Muenster, Germany,
 * info@52north.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; even without the implied WARRANTY OF MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundations web page, http://www.fsf.org.
 *
 ******************************************************************************/
/* $Id: PDPResponse.java 1149 2008-08-05 11:48:03Z MarkoReiprecht $ */
package org.n52.security.decision;

import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;


/**
 * This class represents a request from a AuthzInterceptor to a PDPProxy.
 * PDPResponses are collected in a PDPRequestCollection.
 * A Request can hold several target.
 *
 * @author Bastian Schaeffer
 * @author Marko Reiprecht
 * @version $Revision: 1149 $
 */
public class PDPResponse implements Serializable {

    // ------------------------------ FIELDS ------------------------------

    /**
     * Says that a request is denied.
     */
    public static final int DECISION_DENY = 1;
    /**
     * Says that a request is permitted.
     */
    public static final int DECISION_PERMIT = 2;
    /**
     * Says that a request is undefined, so no decision is maked.
     */
    public static final int DECISION_UNDEFINED = 0;

    private static final long serialVersionUID = 5376242511879896322L;

    protected int m_decision;
    protected PDPRequest m_request;
    protected List m_obligations = new LinkedList();

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * @param decision the decision of the coresponding request.
     * @param request  the request to wich this is a response.
     * @throws IllegalArgumentException if decision is out of bounds
     * @throws NullPointerException     if request is null
     */
    public PDPResponse(final int decision, final PDPRequest request)
            throws IllegalArgumentException, NullPointerException {
        this(decision, request, Collections.EMPTY_LIST);
    }

    /**
     * @param decision    by the PEP
     * @param request     the original request
     * @param obligations by the PEP
     * @throws IllegalArgumentException if descision out of range
     * @throws NullPointerException     if request or obligation are null
     */
    public PDPResponse(final int decision, final PDPRequest request, final Collection obligations)
            throws IllegalArgumentException, NullPointerException {
        checkDecision(decision);
        if (request == null) {
            throw new NullPointerException("request must not null");
        }
        if (obligations == null) {
            throw new NullPointerException("obligations must not null");
        }

        this.m_decision = decision;
        this.m_request = request;
        this.m_obligations.addAll(obligations);
    }

    /**
     * checks decision value range.
     *
     * @param decision a decision
     * @throws IllegalArgumentException if decision out of range
     * @see #DECISION_PERMIT
     * @see #DECISION_DENY
     * @see #DECISION_UNDEFINED
     */
    protected void checkDecision(final int decision) throws IllegalArgumentException {
        switch (decision) {
            case DECISION_PERMIT:
            case DECISION_DENY:
            case DECISION_UNDEFINED:
                break;
            default:
                throw new IllegalArgumentException("decision out of range");
        }
    }

    // --------------------- GETTER / SETTER METHODS ---------------------

    /**
     * @return Returns the decision.
     */
    public int getDecision() {
        return m_decision;
    }

    /**
     * @return Returns the request.
     */
    public PDPRequest getRequest() {
        return m_request;
    }

    // ------------------------ CANONICAL METHODS ------------------------

    /**
     * Method equals.
     *
     * @param o the o.
     * @return boolean.
     */
    public boolean equals(final Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        final PDPResponse that = (PDPResponse) o;

        if (m_decision != that.m_decision) {
            return false;
        }
        if (!m_obligations.equals(that.m_obligations)) {
            return false;
        }
        if (!m_request.equals(that.m_request)) {
            return false;
        }

        return true;
    }

    /**
     * Method hashCode.
     * @return int.
     */
    public int hashCode() {
        int result;
        result = m_decision;
        result = 31 * result + m_request.hashCode();
        result = 31 * result + m_obligations.hashCode();
        return result;
    }

    /**
     * Method toString.
     * @return String.
     */
    public String toString() {
        final StringBuffer sb = new StringBuffer();
        sb.append("org.n52.security.decision.PDPResponse");
        sb.append("{decision=").append(getDecision());
        sb.append(", obligations=").append(getObligations());
        sb.append(", request=").append(getRequest());
        sb.append('}');
        return sb.toString();
    }


    /**
     * Method getObligations returns the obligations of this PDPResponse object.
     *
     * @return the obligations (type Collection) of this PDPResponse object.
     */
    public Collection getObligations() {
        return Collections.unmodifiableCollection(m_obligations);
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Says if this response has obligations.
     *
     * @return boolean
     */
    public boolean hasObligations() {
        return !m_obligations.isEmpty();
    }

    /**
     * A short cut for {@link #DECISION_PERMIT} == {@link #getDecision()}.
     * The method is provided for faster access.
     *
     * @return true if {@link # DECISION_PERMIT} == {@link #getDecision()}.
     */
    public boolean isPermit() {
        return DECISION_PERMIT == m_decision;
    }

    /**
     * A short cut for {@link #DECISION_DENY} == {@link #getDecision()}.
     * The method is provided for faster access.
     *
     * @return true if {@link # DECISION_DENY} == {@link #getDecision()}.
     */
    public boolean isDeny() {
        return DECISION_DENY == m_decision;
    }

    /**
     * A short cut for {@link #DECISION_UNDEFINED} == {@link #getDecision()}.
     * The method is provided for faster access.
     *
     * @return true if {@link # DECISION_UNDEFINED} == {@link #getDecision()}.
     */
    public boolean isUndefined() {
        return DECISION_UNDEFINED == m_decision;
    }
}